import datetime as dt
import re
import tempfile
from os import PathLike, path, listdir, mkdir
from pathlib import Path
from typing import Union, Optional, Dict

import pandas as pd
import patoolib
import requests
from bs4 import BeautifulSoup
from pandas.tseries.offsets import MonthEnd

from econuy.utils import updates, metadata
from econuy.utils.lstrings import fiscal_url, fiscal_sheets


def get(update_path: Union[str, PathLike, None] = None,
        revise_rows: Union[str, int] = "nodup",
        save_path: Union[str, PathLike, None] = None,
        force_update: bool = False,
        name: Optional[str] = None) -> Dict[str, pd.DataFrame]:
    """Get fiscal data.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    force_update : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window (for fiscal accounts, 25 days).
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Monthly fiscal accounts different aggregations : Dict[str, pd.DataFrame]
        Available aggregations: non-financial public sector, consolidated
        public sector, central government, aggregated public enterprises
        and individual public enterprises.

    """
    update_threshold = 25
    if name is None:
        name = "fiscal"

    if update_path is not None:
        full_update_path = (Path(update_path)
                            / f"{name}_nfps").with_suffix(".csv")
        try:
            modified = dt.datetime.fromtimestamp(
                path.getmtime(full_update_path))
            delta = (dt.datetime.now() - modified).days

            if delta < update_threshold and force_update is False:
                print(f"Fiscal data ({full_update_path}) was modified within "
                      f"{update_threshold} day(s). Skipping download...")
                output = {}
                for meta in fiscal_sheets.values():
                    full_update_path = (Path(update_path)
                                        / f"{name}_"
                                          f"{meta['Name']}").with_suffix(
                        ".csv")
                    delta, previous_data = updates._check_modified(
                        full_update_path)
                    output.update({meta["Name"]: previous_data})
                return output
        except FileNotFoundError:
            pass

    response = requests.get(fiscal_url)
    soup = BeautifulSoup(response.content, "html.parser")
    links = soup.find_all(href=re.compile("\\.rar$"))
    rar = links[0]["href"]
    temp_rar = tempfile.NamedTemporaryFile(suffix=".rar").name
    with open(temp_rar, "wb") as f:
        f.write(requests.get(rar).content)

    with tempfile.TemporaryDirectory() as temp_dir:
        patoolib.extract_archive(temp_rar, outdir=temp_dir, verbosity=-1)
        path_temp = path.join(temp_dir, listdir(temp_dir)[0])

        output = {}
        with pd.ExcelFile(path_temp) as xls:
            for sheet, meta in fiscal_sheets.items():
                data = (pd.read_excel(xls, sheet_name=sheet).
                        dropna(axis=0, thresh=4).dropna(axis=1, thresh=4).
                        transpose().set_index(2, drop=True))
                data.columns = data.iloc[0]
                data = data[data.index.notnull()].rename_axis(None)
                data.index = data.index + MonthEnd(1)
                data.columns = meta["Colnames"]

                if update_path is not None:
                    full_update_path = (Path(update_path)
                                        / f"{name}_"
                                          f"{meta['Name']}").with_suffix(
                        ".csv")
                    delta, previous_data = updates._check_modified(
                        full_update_path)
                    data = updates._revise(new_data=data,
                                           prev_data=previous_data,
                                           revise_rows=revise_rows)
                data = data.apply(pd.to_numeric, errors="coerce")
                metadata._set(
                    data, area="Cuentas fiscales y deuda", currency="UYU",
                    inf_adj="No", unit="Millones", seas_adj="NSA",
                    ts_type="Flujo", cumperiods=1
                )

                if save_path is not None:
                    full_save_path = (Path(save_path)
                                      / f"{name}_"
                                        f"{meta['Name']}").with_suffix(
                        ".csv")
                    if not path.exists(path.dirname(full_save_path)):
                        mkdir(path.dirname(full_save_path))
                    data.to_csv(full_save_path)

                output.update({meta["Name"]: data})

    return output



import datetime as dt
from os import PathLike, path, mkdir
from pathlib import Path
from typing import Union, Optional

import numpy as np
import pandas as pd
import requests
from pandas.tseries.offsets import MonthEnd

from econuy import transform
from econuy.utils import updates, metadata
from econuy.utils.lstrings import reer_url, ar_cpi_url, ar_cpi_payload
from econuy.retrieval import cpi, nxr


def get_official(update_path: Union[str, PathLike, None] = None,
                 revise_rows: Union[str, int] = "nodup",
                 save_path: Union[str, PathLike, None] = None,
                 force_update: bool = False,
                 name: Optional[str] = None) -> pd.DataFrame:
    """Get official real exchange rates from the BCU website.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    force_update : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window (for real exchange rates, 25 days).
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Monthly real exchange rates vs select countries/regions : pd.DataFrame
        Available: global, regional, extraregional, Argentina, Brazil, US.

    """
    update_threshold = 25
    if name is None:
        name = "rxr_official"

    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        delta, previous_data = updates._check_modified(full_update_path)

        if delta < update_threshold and force_update is False:
            print(f"{full_update_path} was modified within {update_threshold} "
                  f"day(s). Skipping download...")
            return previous_data

    raw = pd.read_excel(reer_url, skiprows=8, usecols="B:H", index_col=0)
    proc = raw.dropna(how="any")
    proc.columns = ["Global", "Regional", "Extrarregional",
                    "Argentina", "Brasil", "EEUU"]
    proc.index = pd.to_datetime(proc.index) + MonthEnd(1)

    if update_path is not None:
        proc = updates._revise(new_data=proc, prev_data=previous_data,
                               revise_rows=revise_rows)

    metadata._set(proc, area="Precios y salarios", currency="UYU/Otro",
                  inf_adj="No", unit="2017=100", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        proc.to_csv(full_save_path)

    return proc


def get_custom(update_path: Union[str, PathLike, None] = None,
               revise_rows: Union[str, int] = "nodup",
               save_path: Union[str, PathLike, None] = None,
               force_update: bool = False,
               name: Optional[str] = None) -> pd.DataFrame:
    """Get official real exchange rates from the BCU website.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    force_update : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window (for real exchange rates, 25 days).
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Monthly real exchange rates vs select countries : pd.DataFrame
        Available: Argentina, Brazil, US.

    """
    update_threshold = 25
    if name is None:
        name = "rxr_custom"

    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        delta, previous_data = updates._check_modified(full_update_path)

        if delta < update_threshold and force_update is False:
            print(f"{full_update_path} was modified within {update_threshold}"
                  f" day(s). Skipping download...")
            return previous_data

    url_ = "http://dataservices.imf.org/REST/SDMX_JSON.svc/CompactData/IFS/M."
    url_extra = ".?startPeriod=1970&endPeriod="
    raw = []
    for country in ["US", "BR", "AR"]:
        for indicator in ["PCPI_IX", "ENDA_XDC_USD_RATE"]:
            base_url = (f"{url_}{country}.{indicator}{url_extra}"
                        f"{dt.datetime.now().year}")
            r_json = requests.get(base_url).json()
            data = r_json["CompactData"]["DataSet"]["Series"]["Obs"]
            try:
                data = pd.DataFrame(data)
                data.set_index("@TIME_PERIOD", drop=True, inplace=True)
            except ValueError:
                data = pd.DataFrame(np.nan,
                                    index=pd.date_range(start="1970-01-01",
                                                        end=dt.datetime.now(),
                                                        freq="M"),
                                    columns=[f"{country}.{indicator}"])
            if "@OBS_STATUS" in data.columns:
                data.drop("@OBS_STATUS", inplace=True, axis=1)
            data.index = (pd.to_datetime(data.index, format="%Y-%m")
                          + MonthEnd(1))
            data.columns = [f"{country}.{indicator}"]
            raw.append(data)
    raw = pd.concat(raw, axis=1, sort=True).apply(pd.to_numeric)

    ar_black_xr, ar_cpi = _missing_ar()
    proc = raw.copy()
    proc["AR.PCPI_IX"] = ar_cpi
    ar_black_xr = pd.concat([ar_black_xr, proc["AR.ENDA_XDC_USD_RATE"]],
                            axis=1)
    ar_black_xr[0] = np.where(pd.isna(ar_black_xr[0]),
                              ar_black_xr["AR.ENDA_XDC_USD_RATE"],
                              ar_black_xr[0])
    proc["AR.ENDA_XDC_USD_RATE_black"] = ar_black_xr.iloc[:, 0]
    proc["AR_E_A"] = proc.iloc[:, [5, 6]].mean(axis=1)

    uy_cpi = cpi.get(update_path=update_path, revise_rows=6,
                     save_path=save_path, force_update=False)
    uy_e = nxr.get_monthly(update_path=update_path, revise_rows=6,
                           save_path=save_path,
                           force_update=False).iloc[:, [1]]
    proc = pd.concat([proc, uy_cpi, uy_e], axis=1)
    proc = proc.interpolate(method="linear", limit_area="inside")
    proc = proc.dropna(how="any")
    proc.columns = ["US_P", "US_E", "BR_P", "BR_E", "AR_P", "AR_E",
                    "AR_E_B", "AR_E_A", "UY_P", "UY_E"]

    output = pd.DataFrame()
    output["UY_E_P"] = proc["UY_E"] / proc["UY_P"]
    output["TCR_UY_AR"] = output["UY_E_P"] / proc["AR_E_A"] * proc["AR_P"]
    output["TCR_UY_BR"] = output["UY_E_P"] / proc["BR_E"] * proc["BR_P"]
    output["TCR_UY_US"] = output["UY_E_P"] * proc["US_P"]
    output["TCR_AR_US"] = proc["BR_E"] * proc["US_P"] / proc["BR_P"]
    output["TCR_BR_US"] = proc["AR_E"] * proc["US_P"] / proc["AR_P"]
    output.drop("UY_E_P", axis=1, inplace=True)
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Precios y salarios", currency="UYU/Otro",
                  inf_adj="No", unit="-", seas_adj="NSA",
                  ts_type="-", cumperiods=1)
    output = transform.base_index(output, start_date="2010-01-01",
                                  end_date="2010-12-31", base=100)

    if update_path is not None:
        output = updates._revise(new_data=output, prev_data=previous_data,
                                 revise_rows=revise_rows)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        output.to_csv(full_save_path)

    return output


def _missing_ar():
    """Get Argentina's non-official exchange rate and CPI."""
    today_fmt = dt.datetime.now().strftime("%d-%m-%Y")
    black_url = (f"https://mercados.ambito.com/dolar/informal/"
                 f"historico-general/11-01-2002/{today_fmt}")
    black_r = requests.get(black_url).json()
    black_xr = pd.DataFrame(black_r)
    black_xr.set_index(0, drop=True, inplace=True)
    black_xr.drop("Fecha", inplace=True)
    black_xr = black_xr.replace(",", ".", regex=True).apply(pd.to_numeric)
    black_xr.index = pd.to_datetime(black_xr.index, format="%d-%m-%Y")
    black_xr = black_xr.mean(axis=1).to_frame().sort_index()
    black_xr = black_xr.resample("M").mean()

    cpi_r = requests.get(ar_cpi_url, params=ar_cpi_payload)
    cpi_ar = pd.read_html(cpi_r.content)[0]
    cpi_ar.set_index("Fecha", drop=True, inplace=True)
    cpi_ar.index = pd.to_datetime(cpi_ar.index, format="%d/%m/%Y")
    cpi_ar.columns = ["nivel"]
    cpi_ar = cpi_ar.divide(10)

    ps_url = "http://www.inflacionverdadera.com/Argentina_inflation.xls"
    cpi_ps = pd.read_excel(ps_url)
    cpi_ps.set_index("date", drop=True, inplace=True)
    cpi_ps.index = cpi_ps.index + MonthEnd(1)
    cpi_ps = cpi_ps.loc[(cpi_ps.index >= "2006-12-31") &
                        (cpi_ps.index <= "2016-12-01"), "index"]
    cpi_ps = cpi_ps.to_frame().pct_change(periods=1).multiply(100)
    cpi_ps.columns = ["nivel"]

    cpi_all = (cpi_ar.append(cpi_ps).reset_index().
               drop_duplicates(subset="index", keep="last").
               set_index("index", drop=True).sort_index())

    cpi_all = cpi_all.divide(100).add(1).cumprod()

    return black_xr, cpi_all



import logging
from datetime import date
from os import PathLike, path, makedirs, mkdir
from pathlib import Path
from typing import Union, Optional

import pandas as pd

from econuy import frequent, transform
from econuy.utils import logutil
from econuy.retrieval import (cpi, nxr, fiscal_accounts, national_accounts,
                              labor, rxr, commodity_index, fx_operations)


class Session(object):
    """
    Main class to access download and processing methods.

    Attributes
    ----------
    data_dir : str or os.PathLike, default 'econuy-data'
        Directory indicating where data will be saved to and retrieved from.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    force_update : bool, default False
        Whether to force download even if data was recently modified.
    dataset : pd.DataFrame, default pd.DataFrame(index=[], columns=[])
        Current working dataset.
    log : {str, 0, 1, 2}
        Controls how logging works. ``0``: don't log; ``1``: log to console;
        ``2``: log to console and file with default file; ``str``: log to
        console and file with filename=str
    logger : logging.Logger, default None
        Logger object. For most cases this attribute should be ``None``,
        allowing :attr:`log` to control how logging works.
    inplace : bool, default False
        If True, transformation methods will modify the :attr:`dataset`
        inplace and return the input :class:`Session` instance.

    """

    def __init__(self,
                 data_dir: Union[str, PathLike] = "econuy-data",
                 revise_rows: Union[int, str] = "nodup",
                 force_update: bool = False,
                 dataset: Union[dict, pd.DataFrame] = pd.DataFrame(),
                 log: Union[int, str] = 1,
                 logger: Optional[logging.Logger] = None,
                 inplace: bool = False):
        self.data_dir = data_dir
        self.revise_rows = revise_rows
        self.force_update = force_update
        self.dataset = dataset
        self.log = log
        self.logger = logger
        self.inplace = inplace

        if not path.exists(self.data_dir):
            makedirs(self.data_dir)

        if logger is not None:
            self.log = "custom"
        else:
            if isinstance(log, int) and (log < 0 or log > 2):
                raise ValueError("'log' takes either 0 (don't log info),"
                                 " 1 (log to console), 2 (log to console and"
                                 " default file), or str (log to console and "
                                 "file with filename=str).")
            elif log == 2:
                logfile = Path(self.data_dir) / "info.log"
                log_obj = logutil.setup(file=logfile)
                log_method = f"console and file ({logfile.as_posix()})"
            elif isinstance(log, str) and log != "custom":
                logfile = (Path(self.data_dir) / log).with_suffix(".log")
                log_obj = logutil.setup(file=logfile)
                log_method = f"console and file ({logfile.as_posix()})"
            elif log == 1:
                log_obj = logutil.setup(file=None)
                log_method = "console"
            else:
                log_obj = logutil.setup(null=True)
                log_method = "no logging"
            self.logger = log_obj

            if (isinstance(dataset, pd.DataFrame) and
                    dataset.equals(pd.DataFrame(columns=[], index=[]))):
                dataset_message = "empty dataframe"
            else:
                dataset_message = "custom dataset"
            if isinstance(revise_rows, int):
                revise_method = f"{revise_rows} rows to replace"
            else:
                revise_method = revise_rows
            log_obj.info(f"Created Session object with the "
                         f"following attributes:\n"
                         f"Directory for downloads and updates: {data_dir}\n"
                         f"Update method: {revise_method}\n"
                         f"Dataset: {dataset_message}\n"
                         f"Logging method: {log_method}")

    @logutil.log_getter
    def get(self,
            dataset: str,
            update: bool = True,
            save: bool = True,
            override: Optional[str] = None,
            **kwargs):
        """
        Main download method.

        Parameters
        ----------
        dataset : {'cpi', 'nxr_monthly', 'nxr_daily', 'fiscal', 'naccounts', \
                'labor', 'wages', 'comm_index', 'rxr_custom', 'rxr_official', \
                'fx_ops'}
            Type of data to download.
        update : bool, default True
            Whether to update an existing dataset.
        save : bool, default True
            Whether to save the dataset.
        override : str, default None
            If not None, overrides the saved dataset's default filename.
        **kwargs
            These arguments are passed only to
            :func:`econuy.retrieval.commodity_index.get`. There's two
            options: ``force_update_weights: bool`` and
            ``force_update_prices: bool`` which are self-explanatory. Generally
            you will need to update prices but not weights since the latter are
            annual and take a long time to download.

        Returns
        -------
        :class:`~econuy.session.Session`
            Loads the pd.DataFrame output into the :attr:`dataset`
            attribute.

        Raises
        ------
        ValueError
            If an invalid string is given to the ``dataset`` argument.

        """
        if update is True:
            update_path = Path(self.data_dir)
        else:
            update_path = None
        if save is True:
            save_path = Path(self.data_dir)
        else:
            save_path = None

        if kwargs:
            keywords = ", ".join(f"{k}={v}" for k, v in kwargs.items())
            self.logger.info(f"Used the following keyword "
                             f"arguments: {keywords}")

        if dataset == "cpi" or dataset == "prices":
            output = cpi.get(update_path=update_path,
                             revise_rows=self.revise_rows,
                             save_path=save_path,
                             force_update=self.force_update,
                             name=override)
        elif dataset == "fiscal":
            output = fiscal_accounts.get(update_path=update_path,
                                         revise_rows=self.revise_rows,
                                         save_path=save_path,
                                         force_update=self.force_update,
                                         name=override)
        elif dataset == "nxr_monthly" or dataset == "nxr_m":
            output = nxr.get_monthly(update_path=update_path,
                                     revise_rows=self.revise_rows,
                                     save_path=save_path,
                                     force_update=self.force_update,
                                     name=override)
        elif dataset == "nxr_daily" or dataset == "nxr_d":
            output = nxr.get_daily(update_path=update_path,
                                   save_path=save_path,
                                   name=override)
        elif dataset == "naccounts" or dataset == "na":
            output = national_accounts.get(update_path=update_path,
                                           revise_rows=self.revise_rows,
                                           save_path=save_path,
                                           force_update=self.force_update,
                                           name=override)
        elif dataset == "labor" or dataset == "labour":
            output = labor.get_rates(update_path=update_path,
                                     revise_rows=self.revise_rows,
                                     save_path=save_path,
                                     force_update=self.force_update,
                                     name=override)
        elif dataset == "wages":
            output = labor.get_wages(update_path=update_path,
                                     revise_rows=self.revise_rows,
                                     save_path=save_path,
                                     force_update=self.force_update,
                                     name=override)
        elif dataset == "rxr_custom" or dataset == "rxr-custom":
            output = rxr.get_custom(update_path=update_path,
                                    revise_rows=self.revise_rows,
                                    save_path=save_path,
                                    force_update=self.force_update,
                                    name=override)
        elif dataset == "rxr_official" or dataset == "rxr-official":
            output = rxr.get_official(update_path=update_path,
                                      revise_rows=self.revise_rows,
                                      save_path=save_path,
                                      force_update=self.force_update,
                                      name=override)
        elif dataset == "commodity_index" or dataset == "comm_index":
            output = commodity_index.get(update_path=update_path,
                                         save_path=save_path,
                                         name=override,
                                         **kwargs)
        elif dataset == "fx_ops" or dataset == "fxops":
            output = fx_operations.get(update_path=update_path,
                                       save_path=save_path,
                                       name=override)
        else:
            raise ValueError("Invalid keyword for 'dataset' parameter.")

        self.dataset = output

        return self

    @logutil.log_getter
    def get_frequent(self,
                     dataset: str,
                     update: bool = True,
                     save: bool = True,
                     override: Optional[str] = None,
                     **kwargs):
        """
        Get frequently used datasets.

        Parameters
        ----------
        dataset : {'inflation', 'fiscal', 'naccounts', 'labor', 'real_wages'}
            Type of data to download.
        update : bool, default True
            Whether to update an existing dataset.
        save : bool, default  True
            Whether to save the dataset.
        override : str, default None
            If not None, overrides the saved dataset's default filename.
        **kwargs
            Keyword arguments passed to functions in
            :mod:`econuy.frequent`.

        Returns
        -------
        :class:`~econuy.session.Session`
            Loads the downloaded dataframe into the :attr:`dataset` attribute.

        Raises
        ------
        ValueError
            If an invalid string is given to the ``dataset`` argument.

        """
        if update is True:
            update_path = Path(self.data_dir)
        else:
            update_path = None
        if save is True:
            save_path = Path(self.data_dir)
        else:
            save_path = None

        if dataset == "inflation":
            called_args = logutil.get_called_args(frequent.inflation,
                                                  kwargs)
            output = frequent.inflation(update_path=update_path,
                                        save_path=save_path,
                                        name=override)
        elif dataset == "fiscal":
            called_args = logutil.get_called_args(frequent.fiscal,
                                                  kwargs)
            output = frequent.fiscal(update_path=update_path,
                                     save_path=save_path,
                                     name=override,
                                     **kwargs)
        elif dataset == "naccounts" or dataset == "na":
            called_args = logutil.get_called_args(frequent.nat_accounts,
                                                  kwargs)
            output = frequent.nat_accounts(update_path=update_path,
                                           save_path=save_path,
                                           name=override,
                                           **kwargs)
        elif dataset == "labor" or dataset == "labour":
            called_args = logutil.get_called_args(frequent.labor_rate_people,
                                                  kwargs)
            output = frequent.labor_rate_people(update_path=update_path,
                                                save_path=save_path,
                                                name=override,
                                                **kwargs)
        elif dataset == "wages" or dataset == "real_wages":
            called_args = logutil.get_called_args(frequent.labor_real_wages,
                                                  kwargs)
            output = frequent.labor_real_wages(update_path=update_path,
                                               save_path=save_path,
                                               name=override,
                                               **kwargs)
        else:
            raise ValueError("Invalid keyword for 'dataset' parameter.")

        self.dataset = output
        self.logger.info(f"Used the following keyword "
                         f"arguments: {called_args}")

        return self

    @logutil.log_transformer
    def resample(self, target: str, operation: str = "sum",
                 interpolation: str = "linear"):
        """
        Resample to target frequencies.

        See Also
        --------
        :func:`~econuy.transform.resample`

        """
        if isinstance(self.dataset, dict):
            output = {}
            for key, value in self.dataset.items():
                table = transform.resample(value, target=target,
                                           operation=operation,
                                           interpolation=interpolation)
                output.update({key: table})
        else:
            output = transform.resample(self.dataset, target=target,
                                        operation=operation,
                                        interpolation=interpolation)
        if self.inplace is True:
            self.dataset = output
            return self
        else:
            return Session(data_dir=self.data_dir,
                           revise_rows=self.revise_rows,
                           force_update=self.force_update,
                           dataset=output,
                           logger=self.logger,
                           inplace=self.inplace)

    @logutil.log_transformer
    def chg_diff(self, operation: str = "chg", period_op: str = "last"):
        """
        Calculate pct change or difference.

        See Also
        --------
        :func:`~econuy.transform.chg_diff`

        """
        if isinstance(self.dataset, dict):
            output = {}
            for key, value in self.dataset.items():
                table = transform.chg_diff(value, operation=operation,
                                           period_op=period_op)
                output.update({key: table})
        else:
            output = transform.chg_diff(self.dataset, operation=operation,
                                        period_op=period_op)
        if self.inplace is True:
            self.dataset = output
            return self
        else:
            return Session(data_dir=self.data_dir,
                           revise_rows=self.revise_rows,
                           force_update=self.force_update,
                           dataset=output,
                           logger=self.logger,
                           inplace=self.inplace)

    @logutil.log_transformer
    def decompose(self, flavor: str = "both",
                  trading: bool = True, outlier: bool = True,
                  x13_binary: Union[str, PathLike] = "search",
                  search_parents: int = 1):
        """
        Use `X-13 ARIMA <https://www.census.gov/srd/www/x13as/>`_ to
        decompose time series.

        Parameters
        ----------
        flavor : {'both', 'trend', 'seas'}
            Whether to get the trend component, the seasonally adjusted series
            or both.
        trading : bool, default True
            Whether to automatically detect trading days.
        outlier : bool, default True
            Whether to automatically detect outliers.
        x13_binary: str or os.PathLike, default 'search'
            Location of the X13 binary. If ``search`` is used, will attempt to
            find the binary in the project structure.
        search_parents: int, default 1
            If ``search`` is chosen for ``x13_binary``, this parameter controls
            how many parent directories to go up before recursively searching
            for the binary.

        Raises
        ------
        ValueError
            If an invalid string is given to the ``flavor`` argument.

        See Also
        --------
        :func:`~econuy.transform.decompose`

        """
        if isinstance(self.dataset, dict):
            output = {}
            for key, value in self.dataset.items():
                full = transform.decompose(value,
                                           trading=trading,
                                           outlier=outlier,
                                           x13_binary=x13_binary,
                                           search_parents=search_parents)
                if flavor == "trend":
                    table = full[0]
                elif flavor == "seas" or type == "seasonal":
                    table = full[1]
                elif flavor == "both":
                    table = full
                else:
                    raise ValueError("'flavor' can be one of 'both', 'trend', "
                                     "or 'seas'.")

                output.update({key: table})
        else:
            full = transform.decompose(self.dataset,
                                       trading=trading,
                                       outlier=outlier,
                                       x13_binary=x13_binary,
                                       search_parents=search_parents)
            if flavor == "trend":
                output = full[0]
            elif flavor == "seas" or type == "seasonal":
                output = full[1]
            elif flavor == "both":
                output = full
            else:
                raise ValueError("'flavor' can be one of 'both', 'trend', or"
                                 "'seas'.")
        if self.inplace is True:
            self.dataset = output
            return self
        else:
            return Session(data_dir=self.data_dir,
                           revise_rows=self.revise_rows,
                           force_update=self.force_update,
                           dataset=output,
                           logger=self.logger,
                           inplace=self.inplace)

    @logutil.log_transformer
    def convert(self, flavor: str, update: Union[str, PathLike, None] = None,
                save: Union[str, PathLike, None] = None, **kwargs):
        """
        Convert to other units.

        Raises
        ------
        ValueError
            If an invalid string is given to the ``flavor`` argument.

        See Also
        --------
        :func:`~econuy.transform.convert_usd`,
        :func:`~econuy.transform.convert_real`,
        :func:`~econuy.transform.convert_gdp`

        """
        if isinstance(self.dataset, dict):
            output = {}
            for key, value in self.dataset.items():
                if flavor == "usd":
                    table = transform.convert_usd(value, update_path=update,
                                                  save_path=save)
                elif flavor == "real":
                    table = transform.convert_real(value, update_path=update,
                                                   save_path=save, **kwargs)
                elif flavor == "pcgdp" or flavor == "gdp":
                    table = transform.convert_gdp(value, update_path=update,
                                                  save_path=save)
                else:
                    raise ValueError("'flavor' can be one of 'usd', 'real', "
                                     "or 'pcgdp'.")

                output.update({key: table})
        else:
            if flavor == "usd":
                output = transform.convert_usd(self.dataset,
                                               update_path=update,
                                               save_path=save)
            elif flavor == "real":
                output = transform.convert_real(self.dataset,
                                                update_path=update,
                                                save_path=save, **kwargs)
            elif flavor == "pcgdp" or flavor == "gdp":
                output = transform.convert_gdp(self.dataset,
                                               update_path=update,
                                               save_path=save)
            else:
                raise ValueError("'flavor' can be one of 'usd', 'real', "
                                 "or 'pcgdp'.")
        if self.inplace is True:
            self.dataset = output
            return self
        else:
            return Session(data_dir=self.data_dir,
                           revise_rows=self.revise_rows,
                           force_update=self.force_update,
                           dataset=output,
                           logger=self.logger,
                           inplace=self.inplace)

    @logutil.log_transformer
    def base_index(self, start_date: Union[str, date],
                   end_date: Union[str, date, None] = None, base: float = 100):
        """
        Scale to a period or range of periods.

        See Also
        --------
        :func:`~econuy.transform.base_index`

        """
        if isinstance(self.dataset, dict):
            output = {}
            for key, value in self.dataset.items():
                table = transform.base_index(value, start_date=start_date,
                                             end_date=end_date, base=base)
                output.update({key: table})
        else:
            output = transform.base_index(self.dataset, start_date=start_date,
                                          end_date=end_date, base=base)
        if self.inplace is True:
            self.dataset = output
            return self
        else:
            return Session(data_dir=self.data_dir,
                           revise_rows=self.revise_rows,
                           force_update=self.force_update,
                           dataset=output,
                           logger=self.logger,
                           inplace=self.inplace)

    @logutil.log_transformer
    def rolling(self, periods: Optional[int] = None,
                operation: str = "sum"):
        """
        Calculate rolling averages or sums.

        See Also
        --------
        :func:`~econuy.transform.rolling`

        """
        if isinstance(self.dataset, dict):
            output = {}
            for key, value in self.dataset.items():
                table = transform.rolling(value, periods=periods,
                                          operation=operation)
                output.update({key: table})
        else:
            output = transform.rolling(self.dataset, periods=periods,
                                       operation=operation)
        if self.inplace is True:
            self.dataset = output
            return self
        else:
            return Session(data_dir=self.data_dir,
                           revise_rows=self.revise_rows,
                           force_update=self.force_update,
                           dataset=output,
                           logger=self.logger,
                           inplace=self.inplace)

    def save(self, name: str):
        """Save :attr:`dataset` attribute to a CSV."""
        name = Path(name).with_suffix("").as_posix()

        if isinstance(self.dataset, dict):
            for key, value in self.dataset.items():
                save_path = (Path(self.data_dir) /
                             f"{name}_{key}").with_suffix(".csv")
                if not path.exists(path.dirname(save_path)):
                    mkdir(path.dirname(save_path))
                value.to_csv(save_path)
        else:
            save_path = (Path(self.data_dir) / name).with_suffix(".csv")
            if not path.exists(path.dirname(save_path)):
                mkdir(path.dirname(save_path))
            self.dataset.to_csv(save_path)

        self.logger.info(f"Saved dataset to directory {self.data_dir}.")

    def final(self):
        """Return :attr:`dataset` attribute."""
        self.logger.info(f"Retrieved dataset from Session() object.")

        return self.dataset



import shutil
from os import listdir, remove, path
from typing import Tuple
from pathlib import Path

import pandas as pd
import pytest

from econuy import transform
from econuy.utils import metadata
from econuy.utils.lstrings import fiscal_metadata
from econuy.retrieval import nxr
from econuy.session import Session
from .test_transform import dummy_df

CUR_DIR = path.abspath(path.dirname(__file__))
TEST_DIR = path.join(path.dirname(CUR_DIR), "test-data")


def remove_clutter(avoid: Tuple[str] = ("fx_ff.csv", "fx_spot_ff.csv",
                                        "reserves_chg.csv",
                                        "commodity_weights.csv",
                                        "nxr_daily.csv")):
    [remove(path.join(TEST_DIR, x)) for x in listdir(TEST_DIR)
     if x not in avoid]
    return


def test_prices_inflation():
    remove_clutter()
    session = Session(data_dir=TEST_DIR)
    assert isinstance(session, Session)
    assert isinstance(session.dataset, pd.DataFrame)
    inflation = session.get_frequent(dataset="inflation").dataset
    remove_clutter()
    prices = session.get(dataset="cpi").dataset
    remove_clutter()
    inter = transform.chg_diff(prices, period_op="inter")
    compare = inflation.iloc[:, [1]]
    inter.columns = compare.columns
    assert compare.equals(inter)
    monthly = transform.chg_diff(prices, period_op="last")
    compare = inflation.iloc[:, [2]]
    monthly.columns = compare.columns
    assert compare.equals(monthly)
    trend, seasadj = transform.decompose(prices, trading=True, outlier=False)
    monthly_sa = transform.chg_diff(seasadj)
    compare = inflation.iloc[:, [3]]
    monthly_sa.columns = compare.columns
    assert compare.equals(monthly_sa)
    monthly_trend = transform.chg_diff(trend)
    compare = inflation.iloc[:, [4]]
    monthly_trend.columns = compare.columns
    assert monthly_trend.equals(monthly_trend)
    remove_clutter()


def test_fiscal():
    remove_clutter()
    session = Session(data_dir=TEST_DIR)
    assert isinstance(session, Session)
    assert isinstance(session.dataset, pd.DataFrame)
    fiscal_tfm = session.get_frequent(dataset="fiscal", aggregation="nfps",
                                      fss=True, unit="gdp").dataset
    remove_clutter()
    fiscal_ = session.get(dataset="fiscal").dataset
    nfps = fiscal_["nfps"]
    gc = fiscal_["gc-bps"]
    proc = pd.DataFrame(index=nfps.index)
    proc["Ingresos: SPNF-SPC"] = nfps["Ingresos: SPNF"]
    proc["Egresos: Primarios SPNF-SPC"] = nfps["Egresos: Primarios SPNF"]
    proc["Egresos: Inversiones SPNF-SPC"] = nfps["Egresos: Inversiones"]
    proc["Intereses: SPNF"] = nfps["Intereses: Totales"]
    proc["Egresos: Totales SPNF"] = (proc["Egresos: Primarios SPNF-SPC"]
                                     + proc["Intereses: SPNF"])
    proc["Resultado: Primario intendencias"] = nfps[
        "Resultado: Primario intendencias"
    ]
    proc["Resultado: Primario BSE"] = nfps["Resultado: Primario BSE"]
    proc["Resultado: Primario SPNF"] = nfps["Resultado: Primario SPNF"]
    proc["Resultado: Global SPNF"] = nfps["Resultado: Global SPNF"]

    proc["Ingresos: FSS"] = gc["Ingresos: FSS"]
    proc["Intereses: FSS"] = gc["Intereses: BPS-FSS"]
    proc["Ingresos: SPNF-SPC aj. FSS"] = (proc["Ingresos: SPNF-SPC"]
                                          - proc["Ingresos: FSS"])
    proc["Intereses: SPNF aj. FSS"] = (proc["Intereses: SPNF"]
                                       - proc["Intereses: FSS"])
    proc["Egresos: Totales SPNF aj. FSS"] = (proc["Egresos: Totales SPNF"]
                                             - proc["Intereses: FSS"])
    proc["Resultado: Primario SPNF aj. FSS"] = (
            proc["Resultado: Primario SPNF"]
            - proc["Ingresos: FSS"])
    proc["Resultado: Global SPNF aj. FSS"] = (proc["Resultado: Global SPNF"]
                                              - proc["Ingresos: FSS"]
                                              + proc["Intereses: FSS"])

    cols = fiscal_metadata["nfps"][True]
    compare = proc.loc[:, cols]
    metadata._set(compare, area="Cuentas fiscales y deuda",
                  currency="UYU", inf_adj="No", unit="No",
                  seas_adj="NSA", ts_type="Flujo", cumperiods=1)
    compare_gdp = transform.rolling(compare, periods=12, operation="sum")
    compare_gdp = transform.convert_gdp(compare_gdp)
    compare_gdp.columns = fiscal_tfm.columns
    assert compare_gdp.equals(fiscal_tfm)
    remove_clutter()
    fiscal_tfm = session.get_frequent(dataset="fiscal", aggregation="nfps",
                                      fss=True, unit="usd",
                                      seas_adj=None).dataset
    compare_usd = transform.convert_usd(compare)
    compare_usd.columns = fiscal_tfm.columns
    assert compare_usd.equals(fiscal_tfm)
    remove_clutter()
    fiscal_tfm = session.get_frequent(dataset="fiscal", aggregation="nfps",
                                      fss=True, unit="real",
                                      seas_adj=None).dataset
    compare_real = transform.convert_real(compare)
    compare_real.columns = fiscal_tfm.columns
    assert compare_real.equals(fiscal_tfm)
    remove_clutter()
    start_date = "2010-01-31"
    end_date = "2010-12-31"
    fiscal_tfm = session.get_frequent(dataset="fiscal", aggregation="nfps",
                                      fss=True, unit="real_usd", seas_adj=None,
                                      start_date=start_date,
                                      end_date=end_date).dataset
    compare_real_usd = transform.convert_real(compare, start_date=start_date,
                                              end_date=end_date)
    xr = nxr.get_monthly(update_path=None, save_path=None)
    compare_real_usd = compare_real_usd.divide(
        xr[start_date:end_date].mean()[1])
    compare_real_usd.columns = fiscal_tfm.columns
    assert compare_real_usd.equals(fiscal_tfm)
    remove_clutter()
    fiscal_tfm = session.get_frequent(dataset="fiscal", aggregation="nfps",
                                      fss=True, unit=None,
                                      seas_adj="trend").dataset
    compare_trend, compare_sa = transform.decompose(compare, outlier=True,
                                                    trading=True)
    compare_trend.columns = fiscal_tfm.columns
    assert compare_trend.equals(fiscal_tfm)
    remove_clutter()
    fiscal_tfm = session.get_frequent(dataset="fiscal", aggregation="nfps",
                                      fss=True, unit=None, seas_adj=None,
                                      cum=12).dataset
    compare_roll = transform.rolling(compare, periods=12, operation="sum")
    compare_roll.columns = fiscal_tfm.columns
    assert compare_roll.equals(fiscal_tfm)
    remove_clutter()
    with pytest.raises(ValueError):
        session.get_frequent(dataset="fiscal", aggregation="nfps",
                             unit="wrong")
    with pytest.raises(ValueError):
        session.get_frequent(dataset="fiscal", aggregation="nfps",
                             seas_adj="wrong")
    with pytest.raises(ValueError):
        session.get_frequent(dataset="fiscal", aggregation="wrong")
    remove_clutter()


def test_labor():
    remove_clutter()
    session = Session(data_dir=TEST_DIR)
    assert isinstance(session, Session)
    assert isinstance(session.dataset, pd.DataFrame)
    labor_tfm = session.get_frequent(dataset="labor", seas_adj="trend").dataset
    labor_tfm = labor_tfm.iloc[:, [0, 1, 2]]
    remove_clutter()
    labor_ = session.get(dataset="labor").dataset
    labor_trend, labor_sa = transform.decompose(labor_, outlier=True,
                                                trading=True)
    labor_trend.columns = labor_tfm.columns
    assert labor_trend.equals(labor_tfm)
    remove_clutter()
    labor_tfm = session.get_frequent(dataset="labor", seas_adj="seas").dataset
    labor_tfm = labor_tfm.iloc[:, [0, 1, 2]]
    labor_sa.columns = labor_tfm.columns
    assert labor_sa.equals(labor_tfm)
    remove_clutter()
    labor_tfm = session.get_frequent(dataset="labor", seas_adj=None).dataset
    compare = labor_.iloc[:, 0].div(labor_.iloc[:, 1]).round(4)
    compare_2 = labor_tfm.iloc[:, 3].div(labor_tfm.iloc[:, 4]).round(4)
    assert compare.equals(compare_2)
    compare = labor_tfm.iloc[:, 3].mul(labor_.iloc[:, 2]).div(100).round(4)
    assert compare.equals(labor_tfm.iloc[:, 5].round(4))
    remove_clutter()
    with pytest.raises(ValueError):
        session.get_frequent(dataset="labor", seas_adj="wrong")


def test_wages():
    remove_clutter()
    session = Session(data_dir=TEST_DIR)
    assert isinstance(session, Session)
    assert isinstance(session.dataset, pd.DataFrame)
    full_wages = session.get_frequent(dataset="real_wages",
                                      seas_adj="trend").dataset
    wages_tfm = full_wages.iloc[:, [0, 1, 2]]
    remove_clutter()
    wages_ = session.get(dataset="wages").dataset
    wages_trend, wages_sa = transform.decompose(wages_, outlier=False,
                                                trading=True)
    wages_trend = transform.base_index(wages_trend, start_date="2008-07-31")
    wages_trend.columns = wages_tfm.columns
    assert wages_trend.equals(wages_tfm)
    remove_clutter()
    full_wages = session.get_frequent(dataset="wages", seas_adj="seas").dataset
    wages_tfm = full_wages.iloc[:, [0, 1, 2]]
    wages_sa = transform.base_index(wages_sa, start_date="2008-07-31")
    wages_sa.columns = wages_tfm.columns
    assert wages_sa.equals(wages_tfm)
    remove_clutter()
    full_wages = session.get_frequent(dataset="wages", seas_adj=None).dataset
    real_wages = full_wages.iloc[:, [3, 4, 5]]
    compare = transform.convert_real(wages_)
    compare = transform.base_index(compare, start_date="2008-07-31")
    compare.columns = real_wages.columns
    assert real_wages.equals(compare)
    remove_clutter()
    with pytest.raises(ValueError):
        session.get_frequent(dataset="wages", seas_adj="wrong")


def test_naccounts():
    remove_clutter()
    session = Session(data_dir=TEST_DIR)
    assert isinstance(session, Session)
    assert isinstance(session.dataset, pd.DataFrame)
    na_tfm = session.get_frequent(dataset="na", supply=True, real=True,
                                  index=False,
                                  off_seas_adj=False, usd=False, cum=1,
                                  seas_adj=None, variation=None).dataset
    remove_clutter()
    na_ = session.get(dataset="na").dataset
    compare = na_["ind_con_nsa"]
    compare.columns = na_tfm.columns
    assert compare.equals(na_tfm)
    remove_clutter()
    na_tfm = session.get_frequent(dataset="na", supply=True, real=False,
                                  index=False, off_seas_adj=False, usd=False,
                                  cum=1,
                                  seas_adj=None, variation=None).dataset
    compare = na_["ind_cur_nsa"]
    compare.columns = na_tfm.columns
    assert compare.equals(na_tfm)
    remove_clutter()
    na_tfm = session.get_frequent(dataset="na", supply=False, real=True,
                                  index=False, off_seas_adj=False, usd=False,
                                  cum=1,
                                  seas_adj=None, variation=None).dataset
    compare = na_["gas_con_nsa"]
    compare.columns = na_tfm.columns
    assert compare.equals(na_tfm)
    remove_clutter()
    na_tfm = session.get_frequent(dataset="na", supply=False, real=True,
                                  index=False, off_seas_adj=False, usd=True,
                                  cum=4,
                                  seas_adj=None, variation="inter").dataset
    compare_mult = transform.convert_usd(compare)
    compare_mult = transform.rolling(compare_mult, periods=4, operation="sum")
    compare_mult = transform.chg_diff(compare_mult, operation="chg",
                                      period_op="inter")
    compare_mult.columns = na_tfm.columns
    assert compare_mult.equals(na_tfm)
    remove_clutter()
    na_tfm = session.get_frequent(dataset="na", supply=False, real=True,
                                  index=False, off_seas_adj=False, usd=False,
                                  cum=1,
                                  seas_adj="trend", variation=None).dataset
    compare_trend, compare_sa = transform.decompose(compare, trading=True,
                                                    outlier=True)
    compare_trend.columns = na_tfm.columns
    assert compare_trend.equals(na_tfm)
    remove_clutter()
    na_tfm = session.get_frequent(dataset="na", supply=False, real=True,
                                  index=False, off_seas_adj=False, usd=False,
                                  cum=1,
                                  seas_adj="seas", variation=None).dataset
    compare_sa.columns = na_tfm.columns
    assert compare_sa.equals(na_tfm)
    with pytest.raises(KeyError):
        session.get_frequent(dataset="na", supply=False, real=True, index=True,
                             off_seas_adj=True, usd=False, cum=1,
                             seas_adj=None, variation=None)
    with pytest.raises(ValueError):
        session.get_frequent(dataset="na", supply=False, real=True,
                             index=False,
                             off_seas_adj=False, usd=False, cum=1,
                             seas_adj="wrong", variation=None)
    with pytest.raises(ValueError):
        session.get_frequent(dataset="na", supply=False, real=True,
                             index=False,
                             off_seas_adj=False, usd=False, cum=1,
                             seas_adj=None, variation="wrong")
    remove_clutter()


def test_edge():
    remove_clutter()
    session = Session(data_dir=TEST_DIR)
    assert isinstance(session, Session)
    assert isinstance(session.dataset, pd.DataFrame)
    session.get(dataset="cpi", update=False, save=False)
    with pytest.raises(ValueError):
        session.get(dataset="wrong")
    with pytest.raises(ValueError):
        session.get_frequent(dataset="wrong")
    remove_clutter()
    session = Session(data_dir="new_dir")
    assert path.isdir("new_dir")
    shutil.rmtree(session.data_dir)
    Session(data_dir=TEST_DIR).get_frequent(dataset="inflation",
                                            update=False, save=False)
    remove_clutter()


def test_save():
    remove_clutter()
    data = dummy_df(freq="M")
    session = Session(data_dir=TEST_DIR, dataset=data)
    session.save(name="test_save")
    assert path.isfile(Path(TEST_DIR) / "test_save.csv")
    remove_clutter()
    data = dummy_df(freq="M")
    session = Session(data_dir=TEST_DIR, dataset={
        "data1": data, "data2": data})
    session.save(name="test_save")
    assert path.isfile(Path(TEST_DIR) / "test_save_data1.csv")
    assert path.isfile(Path(TEST_DIR) / "test_save_data2.csv")
    remove_clutter()
    session.data_dir = "new_dir"
    session.save(name="test_save")
    assert path.isfile(Path(session.data_dir) / "test_save_data1.csv")
    assert path.isfile(Path(session.data_dir) / "test_save_data2.csv")
    session.dataset = data
    session.save(name="test_save")
    assert path.isfile(Path(session.data_dir) / "test_save.csv")
    remove_clutter()
    shutil.rmtree(session.data_dir)


def test_logging(caplog):
    remove_clutter()
    caplog.clear()
    Session(data_dir=TEST_DIR, log="test")
    assert path.isfile(path.join(TEST_DIR, "test.log"))
    remove_clutter()
    caplog.clear()
    Session(data_dir=TEST_DIR, log=2)
    assert path.isfile(path.join(TEST_DIR, "info.log"))
    remove_clutter()
    caplog.clear()
    with pytest.raises(ValueError):
        Session(data_dir=TEST_DIR, log=5)
    remove_clutter()
    caplog.clear()
    Session(data_dir=TEST_DIR, log=1)
    assert "Logging method: console" in caplog.text
    assert "Logging method: console and file" not in caplog.text
    assert not path.isfile(path.join(TEST_DIR, "info.log"))
    caplog.clear()
    remove_clutter()
    Session(data_dir=TEST_DIR, log=0)
    assert caplog.text == ""



import pytest

from econuy.retrieval import cpi
from .test_session import remove_clutter


def test_revise():
    remove_clutter()
    prices = cpi.get(save_path="test-data")
    prices_alt = cpi.get(update_path="test-data", force_update=True,
                         revise_rows=6)
    assert prices.round(4).equals(prices_alt.round(4))
    prices_alt = cpi.get(update_path="test-data", force_update=True,
                         revise_rows="nodup")
    assert prices.round(4).equals(prices_alt.round(4))
    prices_alt = cpi.get(update_path="test-data", force_update=True,
                         revise_rows="auto")
    assert prices.round(4).equals(prices_alt.round(4))
    with pytest.raises(ValueError):
        cpi.get(update_path="test-data", force_update=True,
                revise_rows="wrong")
    remove_clutter()



import datetime as dt
from os import PathLike, path, mkdir
from pathlib import Path
from typing import Union, Optional, Dict

import pandas as pd
from pandas.tseries.offsets import MonthEnd

from econuy import transform
from econuy.utils import updates, metadata
from econuy.utils.lstrings import nat_accounts_metadata


def get(update_path: Union[str, PathLike, None] = None,
        revise_rows: Union[str, int] = "nodup",
        save_path: Union[str, PathLike, None] = None,
        force_update: bool = False,
        name: Optional[str] = None) -> Dict[str, pd.DataFrame]:
    """Get national accounts data.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    force_update : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window (for national accounts, 85 days).
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Quarterly national accounts : Dict[str, pd.DataFrame]
        Each dataframe corresponds to a national accounts table.

    """
    update_threshold = 80
    if name is None:
        name = "naccounts"

    parsed_excels = {}
    for file, meta in nat_accounts_metadata.items():

        if update_path is not None:
            full_update_path = (Path(update_path) /
                                f"{name}_{meta['Name']}").with_suffix(
                ".csv")
            delta, previous_data = updates._check_modified(full_update_path)

            if delta < update_threshold and force_update is False:
                print(f"{full_update_path}.csv was modified within"
                      f" {update_threshold} day(s). Skipping download...")
                parsed_excels.update({meta["Name"]: previous_data})
                continue

        raw = pd.read_excel(file, skiprows=9, nrows=meta["Rows"])
        proc = (raw.drop(columns=["Unnamed: 0"]).
                dropna(axis=0, how="all").dropna(axis=1, how="all"))
        proc = proc.transpose()
        proc.columns = meta["Colnames"]
        proc.drop(["Unnamed: 1"], inplace=True)

        _fix_dates(proc)

        if meta["Unit"] == "Miles":
            proc = proc.divide(1000)
            unit_ = "Millones"
        else:
            unit_ = meta["Unit"]
        if update_path is not None:
            proc = updates._revise(new_data=proc, prev_data=previous_data,
                                   revise_rows=revise_rows)
        proc = proc.apply(pd.to_numeric, errors="coerce")

        metadata._set(proc, area="Actividad económica", currency="UYU",
                      inf_adj=meta["Inf. Adj."],
                      unit=unit_,
                      seas_adj=meta["Seas"], ts_type="Flujo",
                      cumperiods=1)

        if save_path is not None:
            full_save_path = (Path(save_path) /
                              f"{name}_{meta['Name']}").with_suffix(".csv")
            if not path.exists(path.dirname(full_save_path)):
                mkdir(path.dirname(full_save_path))
            proc.to_csv(full_save_path)

        parsed_excels.update({meta["Name"]: proc})

    return parsed_excels


def _fix_dates(df):
    """Cleanup dates inplace in BCU national accounts files."""
    df.index = df.index.str.replace("*", "")
    df.index = df.index.str.replace(r"\bI \b", "3-", regex=True)
    df.index = df.index.str.replace(r"\bII \b", "6-", regex=True)
    df.index = df.index.str.replace(r"\bIII \b", "9-", regex=True)
    df.index = df.index.str.replace(r"\bIV \b", "12-", regex=True)
    df.index = pd.to_datetime(df.index, format="%m-%Y") + MonthEnd(1)


def _lin_gdp(update_path: Union[str, PathLike, None] = None,
             save_path: Union[str, PathLike, None] = None,
             force_update: bool = False):
    """Get nominal GDP data in UYU and USD with forecasts.

    Update nominal GDP data for use in the `convert.convert_gdp()` function.
    Get IMF forecasts for year of last available data point and the next
    year (for example, if the last period available at the BCU website is
    september 2019, fetch forecasts for 2019 and 2020).

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    force_update : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window (for national accounts, 80 days).

    Returns
    -------
    output : Pandas dataframe
        Quarterly GDP in UYU and USD with 1 year forecasts.

    """
    update_threshold = 80
    name = "lin_gdp"

    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        delta, previous_data = updates._check_modified(full_update_path)

        if delta < update_threshold and force_update is False:
            print(f"{full_update_path} was modified within {update_threshold} "
                  f"day(s). Skipping download...")
            return previous_data

    data_uyu = get(update_path=update_path, revise_rows=4, save_path=save_path,
                   force_update=False)["gdp_cur_nsa"]
    data_uyu = transform.rolling(data_uyu, periods=4, operation="sum")
    data_usd = transform.convert_usd(data_uyu)

    data = [data_uyu, data_usd]
    last_year = data_uyu.index.max().year
    if data_uyu.index.max().month == 12:
        last_year += 1

    results = []
    for table, gdp in zip(["NGDP", "NGDPD"], data):
        table_url = (f"https://www.imf.org/external/pubs/ft/weo/2019/02/weodat"
                     f"a/weorept.aspx?sy={last_year - 1}&ey={last_year + 1}"
                     f"&scsm=1&ssd=1&sort=country&ds=.&br=1&pr1.x=27&pr1.y=9&c"
                     f"=298&s={table}&grp=0&a=")
        imf_data = pd.to_numeric(pd.read_html(table_url)[4].iloc[2, [5, 6, 7]])
        imf_data = imf_data.reset_index(drop=True)
        fcast = (gdp.loc[[dt.datetime(last_year - 1, 12, 31)]].
                 multiply(imf_data.iloc[1]).divide(imf_data.iloc[0]))
        fcast = fcast.rename(index={dt.datetime(last_year - 1, 12, 31):
                                    dt.datetime(last_year, 12, 31)})
        next_fcast = (gdp.loc[[dt.datetime(last_year - 1, 12, 31)]].
                      multiply(imf_data.iloc[2]).divide(imf_data.iloc[0]))
        next_fcast = next_fcast.rename(
            index={dt.datetime(last_year - 1, 12, 31):
                   dt.datetime(last_year + 1, 12, 31)}
        )
        fcast = fcast.append(next_fcast)
        gdp = gdp.append(fcast)
        results.append(gdp)

    output = pd.concat(results, axis=1)
    output = output.resample("Q-DEC").interpolate("linear")

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        output.to_csv(full_save_path)

    return output



from os import PathLike, mkdir, path
from pathlib import Path
from typing import Union, Optional

import pandas as pd
from pandas.tseries.offsets import MonthEnd

from econuy.utils import updates, metadata
from econuy.utils.lstrings import cpi_url


def get(update_path: Union[str, PathLike, None] = None,
        revise_rows: Union[str, int] = "nodup",
        save_path: Union[str, PathLike, None] = None,
        force_update: bool = False,
        name: Optional[str] = None) -> pd.DataFrame:
    """Get CPI data.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    force_update : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window (for consumer prices, 25 days).
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Monthly CPI index : pd.DataFrame

    """
    update_threshold = 25
    if name is None:
        name = "cpi"

    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        delta, previous_data = updates._check_modified(full_update_path)

        if delta < update_threshold and force_update is False:
            print(f"{full_update_path} was modified within "
                  f"{update_threshold} day(s). Skipping download...")
            return previous_data

    cpi_raw = pd.read_excel(cpi_url, skiprows=7).dropna(axis=0, thresh=2)
    cpi = (cpi_raw.drop(["Mensual", "Acum.año", "Acum.12 meses"], axis=1).
           dropna(axis=0, how="all").set_index("Mes y año").rename_axis(None))
    cpi.columns = ["Índice de precios al consumo"]
    cpi.index = cpi.index + MonthEnd(1)

    if update_path is not None:
        cpi = updates._revise(new_data=cpi, prev_data=previous_data,
                              revise_rows=revise_rows)

    cpi = cpi.apply(pd.to_numeric, errors="coerce")
    metadata._set(cpi, area="Precios y salarios", currency="-",
                  inf_adj="No", unit="2010-10-31=100", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        cpi.to_csv(full_save_path)

    return cpi



import datetime as dt
from os import PathLike, mkdir, path
from pathlib import Path
from typing import Union, Optional

import pandas as pd
from pandas.tseries.offsets import MonthEnd

from econuy.utils import updates, metadata
from econuy.utils.lstrings import nxr_url, nxr_daily_url


def get_monthly(update_path: Union[str, PathLike, None] = None,
                revise_rows: Union[str, int] = "nodup",
                save_path: Union[str, PathLike, None] = None,
                force_update: bool = False,
                name: Optional[str] = None) -> pd.DataFrame:
    """Get monthly nominal exchange rate data.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    force_update : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window(for nominal exchange rates, 25 days).
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Monthly nominal exchange rates : pd.DataFrame
        Sell rate, monthly average and end of period.

    """
    update_threshold = 25
    if name is None:
        name = "nxr"

    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        delta, previous_data = updates._check_modified(full_update_path)

        if delta < update_threshold and force_update is False:
            print(f"{full_update_path} was modified within {update_threshold} "
                  f"day(s). Skipping download...")
            return previous_data

    nxr_raw = pd.read_excel(nxr_url, skiprows=4, index_col=0, usecols="A,C,F")
    nxr = nxr_raw.dropna(how="any", axis=0)
    nxr.columns = ["Tipo de cambio venta, fin de período",
                   "Tipo de cambio venta, promedio"]
    nxr.index = nxr.index + MonthEnd(1)
    nxr = nxr.apply(pd.to_numeric, errors="coerce")

    if update_path is not None:
        nxr = updates._revise(new_data=nxr, prev_data=previous_data,
                              revise_rows=revise_rows)

    metadata._set(nxr, area="Precios y salarios", currency="UYU/USD",
                  inf_adj="No", unit="-", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        nxr.to_csv(full_save_path)

    return nxr


def get_daily(update_path: Union[str, PathLike, None] = None,
              save_path: Union[str, PathLike, None] = None,
              name: Optional[str] = None) -> pd.DataFrame:
    """Get daily nominal exchange rate data.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Monthly nominal exchange rates : pd.DataFrame
        Sell rate, monthly average and end of period.

    """
    if name is None:
        name = "nxr_daily"

    start_date = dt.datetime(1999, 12, 31)

    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        prev_data = pd.read_csv(full_update_path, index_col=0,
                                header=list(range(9)),
                                float_precision="high")
        metadata._set(prev_data)
        prev_data.index = pd.to_datetime(prev_data.index)
        start_date = prev_data.index[len(prev_data) - 1]

    today = dt.datetime.now()
    runs = (today - start_date).days // 30
    data = []
    if runs > 0:
        for i in range(1, runs + 1):
            from_ = (start_date + dt.timedelta(days=1)).strftime('%d/%m/%Y')
            to_ = (start_date + dt.timedelta(days=30)).strftime('%d/%m/%Y')
            dates = f"%22FechaDesde%22:%22{from_}%22,%22FechaHasta%22:%22{to_}"
            url = f"{nxr_daily_url}{dates}%22,%22Grupo%22:%222%22}}" + "}"
            try:
                data.append(pd.read_excel(url))
                start_date = dt.datetime.strptime(to_, '%d/%m/%Y')
            except TypeError:
                pass
    from_ = (start_date + dt.timedelta(days=1)).strftime('%d/%m/%Y')
    to_ = dt.datetime.now().strftime('%d/%m/%Y')
    dates = f"%22FechaDesde%22:%22{from_}%22,%22FechaHasta%22:%22{to_}"
    url = f"{nxr_daily_url}{dates}%22,%22Grupo%22:%222%22}}" + "}"
    try:
        data.append(pd.read_excel(url))
        output = pd.concat(data, axis=0)
        output = output.pivot(index="Fecha", columns="Moneda",
                              values="Venta").rename_axis(None)
        output.index = pd.to_datetime(output.index, format="%d/%m/%Y",
                                      errors="coerce")
        output.sort_index(inplace=True)
        output.replace(",", ".", regex=True, inplace=True)
        output.columns = ["Tipo de cambio US$, Cable"]
        output = output.apply(pd.to_numeric, errors="coerce")

        metadata._set(output, area="Precios y salarios", currency="UYU/USD",
                      inf_adj="No", unit="-", seas_adj="NSA",
                      ts_type="-", cumperiods=1)
        output.columns.set_levels(["-"], level=2, inplace=True)

        if update_path is not None:
            output = pd.concat([prev_data, output])

    except TypeError:
        if update_path is not None:
            output = prev_data
        else:
            return pd.DataFrame()

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        output.to_csv(full_save_path)

    return output



import logging
import random
import functools
import inspect
from string import ascii_letters
from pathlib import Path
from typing import Optional


def setup(file: Optional[str] = None, null: bool = False):
    """Setup logging."""
    name = ''.join(random.choice(ascii_letters) for i in range(10))
    log_obj = logging.getLogger(name)

    if null is True:
        log_obj.addHandler(logging.NullHandler())
        return log_obj

    log_obj.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(message)s',
                                  datefmt="%d-%b %T")
    ch.setFormatter(formatter)
    log_obj.addHandler(ch)

    if file is not None:
        file_ext = Path(file).with_suffix(".log")
        fh = logging.FileHandler(file_ext)
        fh.setFormatter(formatter)
        log_obj.addHandler(fh)

    return log_obj


def log_getter(func):
    """
    Decorator for :meth:`~econuy.session.Session.get` and
    :meth:`~econuy.session.Session.get_frequent`.

    """
    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        name = func.__name__
        self.logger.info(f"Getting '{kwargs['dataset']}' "
                         f"with the '{name}' method.")
        return func(self, *args, **kwargs)
    return wrapped


def log_transformer(func):
    """
    Decorator for transformation methods in
    :class:`~econuy.session.Session`.

    """
    @functools.wraps(func)
    def wrapped(self, *func_args, **func_kwargs):
        name = func.__name__
        inspection = inspect.getfullargspec(func)
        kwarg_defaults = dict(zip(inspection[0][-len(inspection[3]):],
                                  inspection[3]))
        if func_kwargs:
            params_kwargs = [(k, v) for k, v in func_kwargs.items()]
            kwarg_defaults.update(params_kwargs)
        keywords = ", ".join(f"{k}={v}" for k, v in kwarg_defaults.items())
        if func_args:
            keywords = ", ".join(func_args) + ", " + keywords
        self.logger.info(f"Applying '{name}' transformation with the following"
                         f" parameters:\n{keywords}")
        return func(self, *func_args, **func_kwargs)
    return wrapped


def get_called_args(func, keywords):
    """
    Helper function for getting arguments passed to
    :meth:`~econuy.session.Session.get_frequent`

    """
    inspection = inspect.getfullargspec(func)
    kwarg_defaults = dict(zip(inspection[0][-len(inspection[3]):],
                              inspection[3]))
    if keywords:
        params_kwargs = [(k, v) for k, v in keywords.items()]
        kwarg_defaults.update(params_kwargs)
    return ", ".join(f"{k}={v}" for k, v in kwarg_defaults.items())



from os import PathLike, path, mkdir
from pathlib import Path
from typing import Union, Optional

import pandas as pd
from pandas.tseries.offsets import MonthEnd

from econuy.utils import updates, metadata
from econuy.utils.lstrings import labor_url, wages1_url, wages2_url


def get_rates(update_path: Union[str, PathLike, None] = None,
              revise_rows: Union[str, int] = "nodup",
              save_path: Union[str, PathLike, None] = None,
              force_update: bool = False,
              name: Optional[str] = None) -> pd.DataFrame:
    """Get labor market data.

    Get monthly labor force participation rate, employment rate (employment to
    working-age population) and unemployment rate.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    force_update : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window (for labor market, 25 days).
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Monthly participation, employment and unemployment rates : pd.DataFrame

    """
    update_threshold = 25
    if name is None:
        name = "labor"

    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        delta, previous_data = updates._check_modified(full_update_path)

        if delta < update_threshold and force_update is False:
            print(f"{full_update_path} was modified within {update_threshold} "
                  f"day(s). Skipping download...")
            return previous_data

    labor_raw = pd.read_excel(labor_url, skiprows=39).dropna(axis=0, thresh=2)
    labor = labor_raw[~labor_raw["Unnamed: 0"].str.contains("-|/|Total",
                                                            regex=True)]
    labor = labor[["Unnamed: 1", "Unnamed: 4", "Unnamed: 7"]]
    labor.index = pd.date_range(start="2006-01-01",
                                periods=len(labor), freq="M")
    labor.columns = ["Tasa de actividad", "Tasa de empleo",
                     "Tasa de desempleo"]

    if update_path is not None:
        labor = updates._revise(new_data=labor, prev_data=previous_data,
                                revise_rows=revise_rows)

    labor = labor.apply(pd.to_numeric, errors="coerce")
    metadata._set(labor, area="Mercado laboral", currency="-",
                  inf_adj="No", unit="Tasa", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        labor.to_csv(full_save_path)

    return labor


def get_wages(update_path: Union[str, PathLike, None] = None,
              revise_rows: Union[str, int] = "nodup",
              save_path: Union[str, PathLike, None] = None,
              force_update: bool = False,
              name: Optional[str] = None) -> pd.DataFrame:
    """Get general, public and private sector wages data

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    force_update : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window (for labor market, 25 days).
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Monthly wages separated by public and private sector : pd.DataFrame

    """
    update_threshold = 25
    if name is None:
        name = "wages"

    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        delta, previous_data = updates._check_modified(full_update_path)

        if delta < update_threshold and force_update is False:
            print(f"{full_update_path} was modified within {update_threshold} "
                  f"day(s). Skipping download...")
            return previous_data

    historical = pd.read_excel(wages1_url, skiprows=8, usecols="A:B")
    historical = historical.dropna(how="any").set_index("Unnamed: 0")
    current = pd.read_excel(wages2_url, skiprows=8, usecols="A,C:D")
    current = current.dropna(how="any").set_index("Unnamed: 0")
    wages = pd.concat([historical, current], axis=1)
    wages.index = wages.index + MonthEnd(1)
    wages.columns = ["Índice medio de salarios",
                     "Índice medio de salarios privados",
                     "Índice medio de salarios públicos"]

    if update_path is not None:
        wages = updates._revise(new_data=wages, prev_data=previous_data,
                                revise_rows=revise_rows)

    wages = wages.apply(pd.to_numeric, errors="coerce")
    metadata._set(wages, area="Mercado laboral", currency="UYU",
                  inf_adj="No", unit="2008-07-31=100", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        wages.to_csv(full_save_path)

    return wages



import datetime as dt
import tempfile
import zipfile
from io import BytesIO
from os import PathLike, path, mkdir
from pathlib import Path
from typing import Union, Optional

import numpy as np
import pandas as pd
import requests
from pandas.tseries.offsets import YearEnd

from econuy.utils import updates, metadata
from econuy.utils.lstrings import (beef_url, pulp_url, soybean_url,
                                   what_url, imf_url, milk1_url, milk2_url)


def _weights(update_path: Union[str, PathLike, None] = None,
             revise_rows: Union[str, int] = "nodup",
             save_path: Union[str, PathLike, None] = None,
             force_update: bool = False) -> pd.DataFrame:
    """Get commodity export weights for Uruguay.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    force_update : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window (for commodity weights, 85 days).

    Returns
    -------
    Commodity weights : pd.DataFrame
        Export-based weights for relevant commodities to Uruguay.

    """
    update_threshold = 85
    name = "commodity_weights"

    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        delta, previous_data = updates._check_modified(full_update_path,
                                                       multiindex=False)

        if delta < update_threshold and force_update is False:
            print(f"{full_update_path} was modified within {update_threshold} "
                  f"day(s). Skipping download...")
            return previous_data

    raw = []
    prods = "%2C".join(["0011", "011", "01251", "01252", "0176", "022", "041",
                        "042", "043", "2222", "24", "25", "268", "97"])
    base_url = "http://comtrade.un.org/api/get?max=1000&type=C&freq=A&px=S3&ps"
    for year in range(1992, dt.datetime.now().year-1):
        full_url = f"{base_url}={year}&r=all&p=858&rg=1&cc={prods}"
        un_r = requests.get(full_url)
        raw.append(pd.DataFrame(un_r.json()["dataset"]))
    raw = pd.concat(raw, axis=0)

    table = raw.groupby(["period", "cmdDescE"]).sum().reset_index()
    table = table.pivot(index="period", columns="cmdDescE",
                        values="TradeValue")
    table.fillna(0, inplace=True)
    percentage = table.div(table.sum(axis=1), axis=0)
    percentage.index = (pd.to_datetime(percentage.index, format="%Y")
                        + YearEnd(1))
    roll = percentage.rolling(window=3, min_periods=3).mean()
    output = roll.resample("M").bfill()

    beef = ["BOVINE MEAT", "Edible offal of bovine animals, fresh or chilled",
            "Meat and offal (other than liver), of bovine animals, "
            "prepared or preserv", "Edible offal of bovine animals, frozen",
            "Bovine animals, live"]
    output["Beef"] = output[beef].sum(axis=1, min_count=len(beef))
    output.drop(beef, axis=1, inplace=True)
    output.columns = ["Barley", "Wood", "Gold", "Milk", "Pulp",
                      "Rice", "Soybeans", "Wheat", "Wool", "Beef"]

    if update_path is not None:
        output = updates._revise(new_data=output, prev_data=previous_data,
                                 revise_rows=revise_rows)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        output.to_csv(full_save_path)

    return output


def _prices(update_path: Union[str, PathLike, None] = None,
            revise_rows: Union[str, int] = "nodup",
            save_path: Union[str, PathLike, None] = None,
            force_update: bool = False) -> pd.DataFrame:
    """Get commodity prices for Uruguay.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    force_update : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window (for commodity prices, 10 days).

    Returns
    -------
    Commodity prices : pd.DataFrame
        Prices and price indexes of relevant commodities for Uruguay.

    """
    update_threshold = 10
    bushel_conv = 36.74 / 100
    name = "commodity_prices"

    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        delta, previous_data = updates._check_modified(full_update_path,
                                                       multiindex=False)

        if delta < update_threshold and force_update is False:
            print(f"{full_update_path} was modified within {update_threshold} "
                  f"day(s). Skipping download...")
            return previous_data

    raw_beef = (pd.read_excel(beef_url, header=4, index_col=0)
                .dropna(how="all"))
    raw_beef.columns = raw_beef.columns.str.strip()
    proc_beef = raw_beef["Ing. Prom./Ton."].to_frame()
    proc_beef.index = pd.date_range(start="2002-01-04",
                                    periods=len(proc_beef), freq="W-SAT")
    proc_beef["Ing. Prom./Ton."] = np.where(
        proc_beef > np.mean(proc_beef) + np.std(proc_beef) * 2,
        proc_beef / 1000,
        proc_beef,
    )
    beef = proc_beef.resample("M").mean()

    raw_pulp_r = requests.get(pulp_url)
    temp_dir = tempfile.TemporaryDirectory()
    with zipfile.ZipFile(BytesIO(raw_pulp_r.content), "r") as f:
        f.extractall(path=temp_dir.name)
        path_temp = path.join(temp_dir.name, "monthly_values.csv")
        raw_pulp = pd.read_csv(path_temp, sep=";").dropna(how="any")
    proc_pulp = raw_pulp.copy().sort_index(ascending=False)
    proc_pulp.index = pd.date_range(start="1990-01-31",
                                    periods=len(proc_pulp), freq="M")
    proc_pulp.drop(["Label", "Codes"], axis=1, inplace=True)
    pulp = proc_pulp

    soy_wheat = []
    for url in [soybean_url, what_url]:
        raw = pd.read_csv(url, index_col=0)
        proc = (raw["Settle"] * bushel_conv).to_frame()
        proc.index = pd.to_datetime(proc.index, format="%Y-%m-%d")
        proc.sort_index(inplace=True)
        soy_wheat.append(proc.resample("M").mean())
    soybean = soy_wheat[0]
    wheat = soy_wheat[1]

    raw_milk = pd.read_excel(milk1_url, skiprows=13,
                             nrows=dt.datetime.now().year - 2006)
    raw_milk.dropna(how="all", axis=1, inplace=True)
    raw_milk.drop(["Promedio ", "Variación"], axis=1, inplace=True)
    raw_milk.columns = ["Año/Mes"] + list(range(1, 13))
    proc_milk = pd.melt(raw_milk, id_vars=["Año/Mes"])
    proc_milk.sort_values(by=["Año/Mes", "variable"], inplace=True)
    proc_milk.index = pd.date_range(start="2007-01-31",
                                    periods=len(proc_milk), freq="M")
    proc_milk = proc_milk.iloc[:, 2].to_frame()

    prev_milk = pd.read_excel(milk2_url, sheet_name="Dairy Products Prices",
                              index_col=0, usecols="A,D", skiprows=5)
    prev_milk = prev_milk.resample("M").mean()
    eurusd_r = requests.get(
        "http://fx.sauder.ubc.ca/cgi/fxdata",
        params=f"b=USD&c=EUR&rd=&fd=1&fm=1&fy=2001&ld=31&lm=12&ly="
               f"{dt.datetime.now().year}&y=monthly&q=volume&f=html&o=&cu=on"
    )
    eurusd = pd.read_html(eurusd_r.content)[0].drop("MMM YYYY", axis=1)
    eurusd.index = pd.date_range(start="2001-01-31", periods=len(eurusd),
                                 freq="M")
    eurusd = eurusd.reindex(prev_milk.index)
    prev_milk = prev_milk.divide(eurusd.values).multiply(10)
    prev_milk = prev_milk.loc[prev_milk.index < min(proc_milk.index)]
    prev_milk.columns, proc_milk.columns = ["Price"], ["Price"]
    milk = prev_milk.append(proc_milk)

    raw_imf = pd.read_excel(imf_url)
    raw_imf.columns = raw_imf.iloc[0, :]
    proc_imf = raw_imf.iloc[3:, 1:]
    proc_imf.index = pd.date_range(start="1980-01-31",
                                   periods=len(proc_imf), freq="M")
    rice = proc_imf[proc_imf.columns[proc_imf.columns.str.contains("Rice")]]
    wood = proc_imf[proc_imf.columns[
        proc_imf.columns.str.contains("Sawnwood")
    ]]
    wood = wood.mean(axis=1).to_frame()
    wool = proc_imf[proc_imf.columns[proc_imf.columns.str.startswith("Wool")]]
    wool = wool.mean(axis=1).to_frame()
    barley = proc_imf[proc_imf.columns[
        proc_imf.columns.str.startswith("Barley")
    ]]
    gold = proc_imf[proc_imf.columns[
        proc_imf.columns.str.startswith("Gold")
    ]]

    complete = pd.concat([beef, pulp, soybean, milk, rice, wood, wool, barley,
                          gold, wheat], axis=1)
    complete = complete.reindex(beef.index).dropna(thresh=8)
    complete.columns = ["Beef", "Pulp", "Soybeans", "Milk", "Rice", "Wood",
                        "Wool", "Barley", "Gold", "Wheat"]

    if update_path is not None:
        complete = updates._revise(new_data=complete, prev_data=previous_data,
                                   revise_rows=revise_rows)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        complete.to_csv(full_save_path)

    return complete


def get(update_path: Union[str, PathLike, None] = None,
        save_path: Union[str, PathLike, None] = None,
        force_update_prices: bool = True, force_update_weights: bool = False,
        name: Optional[str] = None) -> pd.DataFrame:
    """Get export-weighted commodity price index for Uruguay.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't update.
    force_update_prices : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window for commodity prices.
    force_update_weights : bool, default False
        If ``True``, fetch data and update existing data even if it was
        modified within its update window for commodity weights.
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Monthly export-weighted commodity index : pd.DataFrame
        Export-weighted average of commodity prices relevant to Uruguay.

    """
    if name is None:
        name = "commodity_index"

    prices = _prices(update_path=update_path, revise_rows="nodup",
                     save_path=save_path, force_update=force_update_prices)
    prices = prices.interpolate(method="linear", limit=1).dropna(how="any")
    prices = prices.pct_change(periods=1)
    weights = _weights(update_path=update_path, revise_rows="nodup",
                       save_path=save_path, force_update=force_update_weights)
    weights = weights[prices.columns]
    weights = weights.reindex(prices.index, method="ffill")

    product = pd.DataFrame(prices.values * weights.values,
                           columns=prices.columns, index=prices.index)
    product = product.sum(axis=1).add(1).to_frame().cumprod()
    product.columns = ["Índice de precios de productos primarios"]

    metadata._set(product, area="Sector externo", currency="USD",
                  inf_adj="No", unit="2002-01-31=1", seas_adj="NSA",
                  ts_type="Flujo", cumperiods=1)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        product.to_csv(full_save_path)

    return product



import platform
from datetime import date
from os import PathLike, getcwd, path
from pathlib import Path
from typing import Union, Optional, Tuple
import warnings

import numpy as np
import pandas as pd
from statsmodels.api import tsa
from statsmodels.tools.sm_exceptions import X13Error
from statsmodels.tsa import x13

from econuy.utils import metadata, updates
from econuy.retrieval import cpi, national_accounts, nxr


def convert_usd(df: pd.DataFrame,
                update_path: Union[str, PathLike, None] = None,
                save_path: Union[str, PathLike, None] = None) -> pd.DataFrame:
    """
    Convert dataframe from UYU to USD.

    Convert a dataframe's columns from Uruguayan pesos to US dollars. Call the
    :func:`econuy.retrieval.nxr.get_monthly` function to obtain nominal
    exchange rates, and take into account whether the input dataframe's
    ``Type``, as defined by its multiindex, is flow or stock, in order to `
    choose end of period or monthly average NXR. Also take into account the
    input dataframe's frequency and whether columns represent rolling averages
    or sums.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.

    Returns
    -------
    Input dataframe measured in US dollars : pd.DataFrame

    """
    inferred_freq = pd.infer_freq(df.index)
    nxr_data = nxr.get_monthly(update_path=update_path, revise_rows=6,
                               save_path=save_path, force_update=False)

    if df.columns.get_level_values("Tipo")[0] == "Flujo":
        metadata._set(nxr_data, ts_type="Flujo")
        nxr_freq = resample(nxr_data, target=inferred_freq,
                            operation="average").iloc[:, [0]]
        cum_periods = int(df.columns.get_level_values("Acum. períodos")[0])
        nxr_freq = rolling(nxr_freq, periods=cum_periods,
                           operation="average")

    elif df.columns.get_level_values("Tipo")[0] == "Stock":
        metadata._set(nxr_data, ts_type="Stock")
        nxr_freq = resample(nxr_data, target=inferred_freq,
                            operation="average").iloc[:, [1]]
    else:
        raise ValueError("Dataframe needs to have a valid 'Type'.")

    nxr_to_use = nxr_freq[nxr_freq.index.isin(df.index)].iloc[:, 0]
    converted_df = df.apply(lambda x: x / nxr_to_use)
    metadata._set(converted_df, currency="USD")

    return converted_df


def convert_real(df: pd.DataFrame, start_date: Union[str, date, None] = None,
                 end_date: Union[str, date, None] = None,
                 update_path: Union[str, PathLike, None] = None,
                 save_path: Union[str, PathLike, None] = None) -> pd.DataFrame:
    """
    Convert dataframe to real prices.

    Convert a dataframe's columns to real prices. Call the
    :func:`econuy.retrieval.cpi.get` function to obtain the consumer price
    index. take into account the input dataframe's frequency and whether
    columns represent rolling averages or sums. Allow choosing a single period,
    a range of dates or no period as a base (i.e., period for which the
    average/sum of input dataframe and output dataframe is the same).

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    start_date : str, datetime.date or None, default None
        If set to a date-like string or a date, and ``end_date`` is None, the
        base period will be ``start_date``.
    end_date : str, datetime.date or None, default None
        If ``start_date`` is set, calculate so that the data is in constant
        prices of ``start_date-end_date``.
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.

    Returns
    -------
    Input dataframe measured at constant prices : pd.DataFrame

    """
    inferred_freq = pd.infer_freq(df.index)
    cpi_data = cpi.get(update_path=update_path, revise_rows=6,
                       save_path=save_path,
                       force_update=False)
    metadata._set(cpi_data, ts_type="Flujo")
    cpi_freq = resample(cpi_data, target=inferred_freq,
                        operation="average").iloc[:, [0]]
    cum_periods = int(df.columns.get_level_values("Acum. períodos")[0])
    cpi_freq = rolling(cpi_freq, periods=cum_periods,
                       operation="average")
    cpi_to_use = cpi_freq[cpi_freq.index.isin(df.index)].iloc[:, 0]

    if start_date is None:
        converted_df = df.apply(lambda x:
                                x / cpi_to_use)
        col_text = "Const."
    elif end_date is None:
        converted_df = df.apply(lambda x:
                                x / cpi_to_use * cpi_to_use[start_date])
        col_text = f"Const. {start_date}"
    else:
        converted_df = df.apply(
            lambda x: x / cpi_to_use * cpi_to_use[start_date:end_date].mean()
        )
        col_text = f"Const. {start_date}_{end_date}"

    metadata._set(converted_df, inf_adj=col_text)

    return converted_df


def convert_gdp(df: pd.DataFrame,
                update_path: Union[str, PathLike, None] = None,
                save_path: Union[str, PathLike, None] = None) -> pd.DataFrame:
    """
    Calculate dataframe as percentage of GDP.

    Convert a dataframe's columns to percentage of GDP. Call the
    :func:`econuy.retrieval.national_accounts._lin_gdp` function to obtain UYU
    and USD quarterly GDP series. Take into account the input dataframe's
    currency for chossing UYU or USD GDP. If frequency of input dataframe is
    higher than quarterly, GDP will be upsampled and linear interpolation will
    be performed to complete missing data.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.

    Returns
    -------
    Input dataframe as a percentage of GDP : pd.DataFrame

    Raises
    ------
    ValueError
        If frequency of input dataframe not any of 'M', 'MS', 'Q', 'Q-DEC', 'A'
        or 'A-DEC'.

    """
    inferred_freq = pd.infer_freq(df.index)
    gdp = national_accounts._lin_gdp(update_path=update_path,
                                     save_path=save_path, force_update=False)

    if inferred_freq in ["M", "MS"]:
        gdp = resample(gdp, target=inferred_freq,
                       operation="upsample", interpolation="linear")
    elif inferred_freq in ["Q", "Q-DEC", "A", "A-DEC"]:
        gdp = gdp.resample(inferred_freq, convention="end").asfreq()
    else:
        raise ValueError("Frequency of input dataframe not any of 'M', 'MS', "
                         "'Q', 'Q-DEC', 'A' or 'A-DEC'.")

    if df.columns.get_level_values("Moneda")[0] == "USD":
        gdp = gdp.iloc[:, 1].to_frame()
    else:
        gdp = gdp.iloc[:, 0].to_frame()

    gdp_to_use = gdp[gdp.index.isin(df.index)].iloc[:, 0]
    converted_df = df.apply(lambda x: x / gdp_to_use).multiply(100)

    metadata._set(converted_df, unit="% PBI")

    return converted_df


def resample(df: pd.DataFrame, target: str, operation: str = "sum",
             interpolation: str = "linear") -> pd.DataFrame:
    """
    Wrapper for the `resample method <https://pandas.pydata.org/pandas-docs
    stable/reference/api/pandas.DataFrame.resample.html>`_ in Pandas.

    Resample taking into account dataframe ``Type`` so that stock data is not
    averaged or summed when resampling; only last value of target frequency
    is considered.

    Parameters
    ----------
    df : Pandas dataframe
        Input dataframe.
    target : str
        Target frequency to resample to. See
        `Pandas offset aliases <https://pandas.pydata.org/pandas-docs/stable/
        user_guide/timeseries.html#offset-aliases>`_
    operation : {'sum', 'average', 'upsample'}
        Operation to use for resampling.
    interpolation : bool, default True
        Method to use when missing data are produced as a result of resampling.
        See `Pandas interpolation method <https://pandas.pydata.org/pandas-docs
        /stable/reference/api/pandas.Series.interpolate.html>`_

    Returns
    -------
    Input dataframe at the frequency defined in ``target`` : pd.DataFrame

    Raises
    ------
    ValueError
        If ``operation`` is not one of available options and if the input
        dataframe does not have a ``Type`` level in its column multiindex.

    Warns
    -----
    UserWarning
        If input dataframe has ``-`` as frequency in its metadata, warn and
        set to flow.

    """
    if df.columns.get_level_values("Tipo")[0] == "-":
        warnings.warn("Dataframe has no Type, setting to 'Flujo'", UserWarning)
        df.columns = df.columns.set_levels(["Flujo"], level="Tipo")

    if df.columns.get_level_values("Tipo")[0] == "Flujo":
        if operation == "sum":
            resampled_df = df.resample(target).sum()
        elif operation == "average":
            resampled_df = df.resample(target).mean()
        elif operation == "upsample":
            resampled_df = df.resample(target).asfreq()
            resampled_df = resampled_df.interpolate(method=interpolation)
        else:
            raise ValueError("Only 'sum', 'average' and 'upsample' "
                             "are accepted operations")

        cum_periods = int(df.columns.get_level_values("Acum. períodos")[0])
        if cum_periods != 1:
            input_notna = df.iloc[:, 0].count()
            output_notna = resampled_df.iloc[:, 0].count()
            cum_adj = round(output_notna / input_notna)
            metadata._set(resampled_df,
                          cumperiods=int(cum_periods * cum_adj))

    else:
        resampled_df = df.resample(target, convention="end").asfreq()
        resampled_df = resampled_df.interpolate(method=interpolation)

    metadata._set(resampled_df)

    return resampled_df


def rolling(df: pd.DataFrame, periods: Optional[int] = None,
            operation: str = "sum") -> pd.DataFrame:
    """
    Wrapper for the `rolling method <https://pandas.pydata.org/pandas-docs/
    stable/reference/api/pandas.DataFrame.rolling.html>`_ in Pandas.

    If ``periods`` is ``None``, try to infer the frequency and set ``periods``
    according to the following logic: ``{'A': 1, 'Q-DEC': 4, 'M': 12}``.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    periods : int, default None
        How many periods the window should cover.
    operation : {'sum', 'average'}
        Operation used to calculate rolling windows.

    Returns
    -------
    Input dataframe with rolling windows : pd.DataFrame

    Warns
    -----
    UserWarning
        If the input dataframe is a stock time series.

    """
    pd_frequencies = {"A": 1,
                      "A-DEC": 1,
                      "Q": 4,
                      "Q-DEC": 4,
                      "M": 12}

    window_operation = {
        "sum": lambda x: x.rolling(window=periods,
                                   min_periods=periods).sum(),
        "average": lambda x: x.rolling(window=periods,
                                       min_periods=periods).mean()
    }

    if df.columns.get_level_values("Tipo")[0] == "Stock":
        warnings.warn("Rolling operations shouldn't be "
                      "calculated on stock variables", UserWarning)

    if periods is None:
        inferred_freq = pd.infer_freq(df.index)
        periods = pd_frequencies[inferred_freq]

    rolling_df = df.apply(window_operation[operation])

    metadata._set(rolling_df, cumperiods=periods)

    return rolling_df


def base_index(df: pd.DataFrame, start_date: Union[str, date],
               end_date: Union[str, date, None] = None,
               base: float = 100) -> pd.DataFrame:
    """Rebase all dataframe columns to a date or range of dates.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    start_date : string or datetime.date
        Date to which series will be rebased.
    end_date : string or datetime.date, default None
        If specified, series will be rebased to the average between
        ``start_date`` and ``end_date``.
    base : float, default 100
        Float for which ``start_date`` == ``base`` or average between
        ``start_date`` and ``end_date`` == ``base``.

    Returns
    -------
    Input dataframe with a base period index : pd.DataFrame

    """
    if end_date is None:
        indexed = df.apply(lambda x: x / x[start_date] * base)
        metadata._set(indexed, unit=f"{start_date}={base}")

    else:
        indexed = df.apply(lambda x: x / x[start_date:end_date].mean() * base)
        metadata._set(indexed, unit=f"{start_date}_{end_date}={base}")

    return indexed


# The `_open_and_read` function needs to be monkey-patched to specify the
# encoding or decomposition will fail on Windows
def _new_open_and_read(fname):
    with open(fname, 'r', encoding='utf8') as fin:
        fout = fin.read()
    return fout


x13._open_and_read = _new_open_and_read


def decompose(df: pd.DataFrame, trading: bool = True, outlier: bool = True,
              x13_binary: Union[str, PathLike] = "search",
              search_parents: int = 1) -> Optional[Tuple[pd.DataFrame,
                                                         pd.DataFrame]]:
    """
    Apply X13 decomposition. Return trend and seasonally adjusted dataframes.

    Decompose the series in a Pandas dataframe using the US Census X13
    methodology. Will try different combinations of the ``trading`` and
    ``outlier`` arguments if an X13 error is raised. Requires providing the X13
    binary. Please refer to the README for instructions on where to get this
    binary.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    trading : bool, default True
        Whether to automatically detect trading days.
    outlier : bool, default True
        Whether to automatically detect outliers.
    x13_binary: str or os.PathLike, default 'search'
        Location of the X13 binary. If ``search`` is used, will attempt to find
        the binary in the project structure.
    search_parents: int, default 1
        If ``search`` is chosen for ``x13_binary``, this parameter controls how
        many parent directories to go up before recursively searching for the
        binary.

    Returns
    -------
    Decomposed dataframes : Tuple[pd.DataFrame, pd.DataFrame] or None
        Tuple containing the trend component and the seasonally adjusted
        series.

    Raises
    ------
    ValueError
        If the path provided for the X13 binary does not point to a file.

    """
    if x13_binary == "search":
        search_term = "x13as"
        if platform.system() == "Windows":
            search_term += ".exe"
        binary_path = updates._rsearch(dir_file=getcwd(), n=search_parents,
                                       search_term=search_term)
    elif isinstance(x13_binary, str):
        binary_path = x13_binary
    else:
        binary_path = Path(x13_binary).as_posix()

    if path.isfile(binary_path) is False:
        raise ValueError("X13 binary missing. Please refer to the README "
                         "for instructions on where to get binaries for "
                         "Windows and Unix, and how to compile it for macOS.")

    df_proc = df.copy()
    old_columns = df_proc.columns
    df_proc.columns = df_proc.columns.get_level_values(level=0)
    df_proc.index = pd.to_datetime(df_proc.index, errors="coerce")

    trends = []
    seas_adjs = []
    for column in range(len(df_proc.columns)):
        series = df_proc.iloc[:, column].dropna()
        try:
            decomposition = tsa.x13_arima_analysis(
                series, outlier=outlier, trading=trading, forecast_periods=0,
                x12path=binary_path, prefer_x13=True
            )
            trend = decomposition.trend.reindex(df_proc.index)
            seas_adj = decomposition.seasadj.reindex(df_proc.index)

        except X13Error:
            if outlier is True:
                try:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}' with selected "
                          f"parameters. Trying with outlier=False...")
                    return decompose(df=df, outlier=False)

                except X13Error:
                    try:
                        print(f"X13 error found while processing "
                              f"'{df_proc.columns[column]}' with "
                              f"trading=True. Trying with trading=False...")
                        return decompose(df=df, outlier=False,
                                         trading=False)

                    except X13Error:
                        print(f"X13 error found while processing "
                              f"'{df_proc.columns[column]}'. "
                              f"Filling with nan.")
                        trend = pd.Series(np.nan, index=df_proc.index)
                        seas_adj = pd.Series(np.nan, index=df_proc.index)

            elif trading is True:
                try:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}' "
                          f"with trading=True. Trying with "
                          f"trading=False...")
                    return decompose(df=df, trading=False)

                except X13Error:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}'. Filling with nan.")
                    trend = pd.Series(np.nan, index=df_proc.index)
                    seas_adj = pd.Series(np.nan, index=df_proc.index)

            else:
                try:
                    return decompose(df=df)

                except X13Error:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}'. "
                          f"Filling with nan.")
                    trend = pd.Series(np.nan, index=df_proc.index)
                    seas_adj = pd.Series(np.nan, index=df_proc.index)

        trends.append(trend)
        seas_adjs.append(seas_adj)

    trends = pd.concat(trends, axis=1)
    seas_adjs = pd.concat(seas_adjs, axis=1)

    trends.columns = old_columns
    seas_adjs.columns = old_columns

    metadata._set(trends, seas_adj="Tendencia")

    metadata._set(seas_adjs, seas_adj="SA")

    return trends, seas_adjs


def chg_diff(df: pd.DataFrame, operation: str = "chg",
             period_op: str = "last") -> pd.DataFrame:
    """
    Wrapper for the `pct_change <https://pandas.pydata.org/pandas-docs/stable/
    reference/api/pandas.DataFrame.pct_change.html>`_ and `diff <https://pandas
    .pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.diff.html>`_
    Pandas methods.

    Calculate percentage change or difference for dataframes. The ``period``
    argument takes into account the frequency of the dataframe, i.e.,
    ``inter`` (for interannual) will calculate pct change/differences with
    ``periods=4`` for quarterly frequency, but ``periods=12`` for monthly
    frequency.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    operation : {'chg', 'diff'}
        ``chg`` for percent change or ``diff`` for differences.
    period_op : {'last', 'inter', 'annual'}
        Period with which to calculate change or difference. ``last`` for
        previous period (last month for monthly data), ``inter`` for same
        period last year, ``annual`` for same period last year but taking
        annual averages/sums.

    Returns
    -------
    Percent change or differences dataframe : pd.DataFrame

    Raises
    ------
    ValueError
        If the dataframe is not of frequency ``M`` (month), ``Q`` or
        ``Q-DEC`` (quarter), or ``A`` or ``A-DEC`` (year).

    """
    inferred_freq = pd.infer_freq(df.index)

    type_change = {"last":
                   {"chg": [lambda x: x.pct_change(periods=1),
                            "% variación"],
                    "diff": [lambda x: x.diff(periods=1),
                             "Cambio"]},
                   "inter":
                   {"chg": [lambda x: x.pct_change(periods=last_year),
                            "% variación interanual"],
                    "diff": [lambda x: x.diff(periods=last_year),
                             "Cambio interanual"]},
                   "annual":
                   {"chg": [lambda x: x.pct_change(periods=last_year),
                            "% variación anual"],
                    "diff": [lambda x: x.diff(periods=last_year),
                             "Cambio anual"]}}

    if inferred_freq == "M":
        last_year = 12
    elif inferred_freq == "Q" or inferred_freq == "Q-DEC":
        last_year = 4
    elif inferred_freq == "A" or inferred_freq == "A-DEC":
        last_year = 1
    else:
        raise ValueError("The dataframe needs to have a frequency of M "
                         "(month end), Q (quarter end) or A (year end)")

    if period_op == "annual":

        if df.columns.get_level_values("Tipo")[0] == "Stock":
            output = df.apply(type_change[period_op][operation][0])
        else:
            output = rolling(df, operation="sum")
            output = output.apply(
                type_change[period_op][operation][0])

        metadata._set(output, unit=type_change[period_op][operation][1])

    else:
        output = df.apply(type_change[period_op][operation][0])
        metadata._set(output, unit=type_change[period_op][operation][1])

    if operation == "chg":
        output = output.multiply(100)

    return output



import datetime as dt
from os import path

import numpy as np
import pandas as pd

from econuy.session import Session
from econuy.retrieval import fx_operations, national_accounts
from econuy.utils import metadata
from .test_session import remove_clutter

CUR_DIR = path.abspath(path.dirname(__file__))
TEST_DIR = path.join(path.dirname(CUR_DIR), "test-data")


def test_changes():
    remove_clutter()
    previous_data = pd.read_csv(path.join(TEST_DIR, "reserves_chg.csv"),
                                index_col=0, header=list(range(9)))
    metadata._set(previous_data)
    res = fx_operations._reserves_changes(
        update_path=TEST_DIR, name=None, save_path=TEST_DIR)
    previous_data.index = pd.to_datetime(previous_data.index)
    compare = res.loc[previous_data.index].round(4)
    compare.columns = previous_data.columns
    assert compare.equals(previous_data.round(4))
    remove_clutter()


def test_ff():
    remove_clutter()
    previous_data = pd.read_csv(path.join(TEST_DIR, "fx_ff.csv"),
                                index_col=0, header=list(range(9)))
    metadata._set(previous_data)
    compare = previous_data.iloc[0:-30]
    ff = fx_operations._futures_forwards(
        update_path=TEST_DIR, name=None, save_path=TEST_DIR)
    assert len(ff) > len(compare)
    remove_clutter()


def test_ops():
    remove_clutter()
    session = Session(data_dir=TEST_DIR)
    ops = session.get(dataset="fx_ops").dataset
    assert isinstance(ops, pd.DataFrame)
    remove_clutter()


def test_rxr_official():
    remove_clutter()
    session = Session(data_dir=TEST_DIR)
    tcr = session.get(dataset="rxr_official").dataset
    assert isinstance(tcr, pd.DataFrame)
    assert tcr.index[0] == dt.date(2000, 1, 31)
    remove_clutter()


def test_rxr_custom():
    remove_clutter()
    session = Session(data_dir=TEST_DIR)
    tcr = session.get(dataset="rxr_custom").dataset
    assert isinstance(tcr, pd.DataFrame)
    assert tcr.index[0] == dt.date(1979, 12, 31)
    assert len(tcr.columns) == 5
    avs = tcr.loc[(tcr.index >= "2010-01-01") &
                  (tcr.index <= "2010-12-31")].mean().values.round(1)
    arr = np.array([100]*5, dtype="float64")
    assert np.all(avs == arr)
    remove_clutter()


def test_comm_index():
    remove_clutter()
    session = Session(data_dir=TEST_DIR)
    comm = session.get(dataset="comm_index").dataset
    assert isinstance(comm, pd.DataFrame)
    assert comm.index[0] == dt.date(2002, 1, 31)
    assert comm.iloc[0][0] == 1
    remove_clutter()


def test_lin():
    remove_clutter()
    lin = national_accounts._lin_gdp(
        update_path="test-data", save_path="test-data")
    assert isinstance(lin, pd.DataFrame)
    assert (sorted(lin.columns.get_level_values("Moneda"))
            == sorted(["UYU", "USD"]))
    remove_clutter()


def test_nxr_daily():
    remove_clutter()
    previous_data = pd.read_csv(path.join(TEST_DIR, "nxr_daily.csv"),
                                index_col=0, header=list(range(9)))
    metadata._set(previous_data)
    previous_data.index = pd.to_datetime(previous_data.index)
    session = Session(data_dir=TEST_DIR)
    nxr = session.get(dataset="nxr_daily").dataset
    compare = nxr.loc[previous_data.index].round(4)
    compare.columns = previous_data.columns
    assert compare.equals(previous_data.round(4))
    remove_clutter()


def test_nxr_monthly():
    remove_clutter()
    session = Session(data_dir=TEST_DIR)
    nxr = session.get(dataset="nxr_m").dataset
    assert len(nxr.columns) == 2
    assert isinstance(nxr.index[0], pd._libs.tslibs.timestamps.Timestamp)
    remove_clutter()



import datetime as dt
from os import path, PathLike
from pathlib import Path
from typing import Union

import pandas as pd

from econuy.utils import metadata


def _check_modified(data_path: Union[str, PathLike], multiindex=True):
    """Check how long it has been since the file in path was modified."""
    try:
        modified_time = dt.datetime.fromtimestamp(path.getmtime(data_path))
        delta = (dt.datetime.now() - modified_time).days
        if multiindex is True:
            previous_data = pd.read_csv(data_path, index_col=0,
                                        header=list(range(9)),
                                        float_precision="high")
            metadata._set(previous_data)
        else:
            previous_data = pd.read_csv(data_path, index_col=0)
        previous_data.index = pd.to_datetime(previous_data.index)
    except FileNotFoundError:
        print(f"{data_path} does not exist. No data will be updated")
        delta = 9999
        previous_data = pd.DataFrame()

    return delta, previous_data


def _revise(new_data: pd.DataFrame, prev_data: pd.DataFrame,
            revise_rows: Union[int, str]):
    """Replace n rows of data at the end of a dataframe with new data."""
    if len(prev_data) == 0:
        return new_data
    frequency = pd.infer_freq(prev_data.index)
    freq_table = {"A": 3, "A-DEC": 3, "Q": 4, "Q-DEC": 4, "M": 12}
    new_data = new_data.apply(pd.to_numeric, errors="coerce")

    if isinstance(revise_rows, str) and revise_rows in "noduplicate":
        prev_data.columns = new_data.columns
        updated = prev_data.append(new_data)
        updated = updated.loc[~updated.index.duplicated(keep="last")]
        updated.sort_index(inplace=True)
        return updated

    elif isinstance(revise_rows, str) and revise_rows in "automatic":
        try:
            revise_rows = freq_table[frequency]
        except KeyError:
            revise_rows = 12
            if len(prev_data) <= 12 or len(new_data) <= 12:
                revise_rows = 3

    elif isinstance(revise_rows, int):
        revise_rows = revise_rows
    else:
        raise ValueError("`revise_rows` accepts int, 'nodup' or 'auto'")

    non_revised = prev_data[:len(prev_data)-revise_rows]
    revised = new_data[len(prev_data)-revise_rows:]
    non_revised.columns = new_data.columns
    updated = non_revised.append(revised, sort=False)

    return updated


def _rsearch(dir_file: Union[str, PathLike], search_term: str, n: int = 2):
    """Recursively search for a file starting from the n-parent folder of
    a supplied path."""
    i = 0
    while i < n:
        i += 1
        dir_file = path.dirname(dir_file)
    try:
        final_path = ([x for x in Path(dir_file).rglob(search_term)][0]
                      .absolute().as_posix())
    except IndexError:
        final_path = True
    return final_path



from datetime import date
from os import PathLike, mkdir, path
from pathlib import Path
from typing import Union, Optional

import pandas as pd

from econuy import transform
from econuy.utils import metadata
from econuy.utils.lstrings import fiscal_metadata, wap_url
from econuy.retrieval import (nxr, national_accounts, cpi,
                              fiscal_accounts, labor)


def inflation(update_path: Union[str, PathLike, None] = None,
              save_path: Union[str, PathLike, None] = None,
              name: Optional[str] = None) -> pd.DataFrame:
    """
    Get common inflation measures.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Prices measures : pd.DataFrame
        Columns: CPI index, annual inflation, monthly inflation, seasonally
        adjusted monthly inflation and trend monthly inflation.

    """
    if name is None:
        name = "tfm_prices"
    data = cpi.get(update_path=update_path, revise_rows=6,
                   save_path=save_path, force_update=False)
    interannual = transform.chg_diff(data, operation="chg", period_op="inter")
    monthly = transform.chg_diff(data, operation="chg", period_op="last")
    trend, seasadj = transform.decompose(data, trading=True, outlier=False)
    monthly_sa = transform.chg_diff(seasadj, operation="chg",
                                    period_op="last")
    monthly_trend = transform.chg_diff(trend, operation="chg",
                                       period_op="last")
    output = pd.concat([data, interannual, monthly,
                        monthly_sa, monthly_trend], axis=1)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        output.to_csv(full_save_path)

    return output


def fiscal(aggregation: str = "gps", fss: bool = True,
           unit: Union[str, None] = "gdp",
           start_date: Union[str, date, None] = None,
           end_date: Union[str, date, None] = None, cum: int = 1,
           seas_adj: Union[str, None] = None,
           update_path: Union[str, PathLike, None] = None,
           save_path: Union[str, PathLike, None] = None,
           name: Optional[str] = None) -> pd.DataFrame:
    """
    Get fiscal accounts data.

    Allow choosing government aggregation, whether to exclude the FSS
    (Fideicomiso  de la Seguridad Social, Social Security Trust Fund), the unit
    (UYU, real UYU, USD, real USD or percent of GDP), periods to accumuldate
    for rolling sums and seasonal adjustment.

    Parameters
    ----------
    aggregation : {'gps', 'nfps', 'gc'}
        Government aggregation. Can be ``gps`` (consolidated public sector),
        ``nfps`` (non-financial public sector) or ``gc`` (central government).
    fss : bool, default True
        If ``True``, exclude the `FSS's <https://www.impo.com.uy/bases/decretos
        /71-2018/25>`_ income from gov't revenues and the FSS's
        interest revenues from gov't interest payments.
    unit : {'gdp', 'usd', 'real', 'real_usd'}
        Unit in which data should be expressed. Possible values are ``real``,
        ``usd``, ``real_usd`` and ``gdp``. If ``None`` or another string is
        set, no unit calculations will be performed, rendering the data as is
        (current UYU).
    start_date : str, datetime.date or None, default None
        If ``unit`` is set to ``real`` or ``real_usd``, this parameter and
        ``end_date`` control how deflation is calculated.
    end_date :
        If ``unit`` is set to ``real`` or ``real_usd``, this parameter and
        ``start_date`` control how deflation is calculated.
    cum : int, default 1
        How many periods to accumulate for rolling sums.
    seas_adj : {None, 'trend', 'seas'}
        Whether to seasonally adjust.
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Fiscal aggregation : pd.DataFrame

    Raises
    ------
    ValueError
        If ``seas_adj``, ``unit`` or ``aggregation`` are given an invalid
        keywords.

    """
    if name is None:
        name = "tfm_fiscal"

    if seas_adj not in ["trend", "seas", None]:
        raise ValueError("'seas_adj' can be 'trend', 'seas' or None.")
    if unit not in ["gdp", "usd", "real", "real_usd", None]:
        raise ValueError("'unit' can be 'gdp', 'usd', 'real', 'real_usd' or"
                         " None.")
    if aggregation not in ["gps", "nfps", "gc"]:
        raise ValueError("'aggregation' can be 'gps', 'nfps' or 'gc'.")

    data = fiscal_accounts.get(update_path=update_path, revise_rows=12,
                               save_path=save_path, force_update=False)
    gps = data["gps"]
    nfps = data["nfps"]
    gc = data["gc-bps"]

    proc = pd.DataFrame(index=gps.index)
    proc["Ingresos: SPNF-SPC"] = nfps["Ingresos: SPNF"]
    proc["Ingresos: GC-BPS"] = gc["Ingresos: GC-BPS"]
    proc["Egresos: Primarios SPNF-SPC"] = nfps["Egresos: Primarios SPNF"]
    proc["Egresos: Totales GC-BPS"] = gc["Egresos: GC-BPS"]
    proc["Egresos: Inversiones SPNF-SPC"] = nfps["Egresos: Inversiones"]
    proc["Egresos: Inversiones GC-BPS"] = gc["Egresos: Inversión"]
    proc["Intereses: SPNF"] = nfps["Intereses: Totales"]
    proc["Intereses: BCU"] = gps["Intereses: BCU"]
    proc["Intereses: SPC"] = proc["Intereses: SPNF"] + proc["Intereses: BCU"]
    proc["Intereses: GC-BPS"] = gc["Intereses: Total"]
    proc["Egresos: Totales SPNF"] = (proc["Egresos: Primarios SPNF-SPC"]
                                     + proc["Intereses: SPNF"])
    proc["Egresos: Totales SPC"] = (proc["Egresos: Totales SPNF"]
                                    + proc["Intereses: BCU"])
    proc["Egresos: Primarios GC-BPS"] = (proc["Egresos: Totales GC-BPS"]
                                         - proc["Intereses: GC-BPS"])
    proc["Resultado: Primario intendencias"] = nfps[
        "Resultado: Primario intendencias"
    ]
    proc["Resultado: Primario BSE"] = nfps["Resultado: Primario BSE"]
    proc["Resultado: Primario BCU"] = gps["Resultado: Primario BCU"]
    proc["Resultado: Primario SPNF"] = nfps["Resultado: Primario SPNF"]
    proc["Resultado: Global SPNF"] = nfps["Resultado: Global SPNF"]
    proc["Resultado: Primario SPC"] = gps["Resultado: Primario SPC"]
    proc["Resultado: Global SPC"] = gps["Resultado: Global SPC"]
    proc["Resultado: Primario GC-BPS"] = (proc["Ingresos: GC-BPS"]
                                          - proc["Egresos: Primarios GC-BPS"])
    proc["Resultado: Global GC-BPS"] = gc["Resultado: Global GC-BPS"]

    proc["Ingresos: FSS"] = gc["Ingresos: FSS"]
    proc["Intereses: FSS"] = gc["Intereses: BPS-FSS"]
    proc["Ingresos: SPNF-SPC aj. FSS"] = (proc["Ingresos: SPNF-SPC"]
                                          - proc["Ingresos: FSS"])
    proc["Ingresos: GC-BPS aj. FSS"] = (proc["Ingresos: GC-BPS"]
                                        - proc["Ingresos: FSS"])
    proc["Intereses: SPNF aj. FSS"] = (proc["Intereses: SPNF"]
                                       - proc["Intereses: FSS"])
    proc["Intereses: SPC aj. FSS"] = (proc["Intereses: SPC"]
                                      - proc["Intereses: FSS"])
    proc["Intereses: GC-BPS aj. FSS"] = (proc["Intereses: GC-BPS"]
                                         - proc["Intereses: FSS"])
    proc["Egresos: Totales SPNF aj. FSS"] = (proc["Egresos: Totales SPNF"]
                                             - proc["Intereses: FSS"])
    proc["Egresos: Totales SPC aj. FSS"] = (proc["Egresos: Totales SPC"]
                                            - proc["Intereses: FSS"])
    proc["Egresos: Totales GC-BPS aj. FSS"] = (proc["Egresos: Totales GC-BPS"]
                                               - proc["Intereses: FSS"])
    proc["Resultado: Primario SPNF aj. FSS"] = (
        proc["Resultado: Primario SPNF"]
        - proc["Ingresos: FSS"])
    proc["Resultado: Global SPNF aj. FSS"] = (proc["Resultado: Global SPNF"]
                                              - proc["Ingresos: FSS"]
                                              + proc["Intereses: FSS"])
    proc["Resultado: Primario SPC aj. FSS"] = (proc["Resultado: Primario SPC"]
                                               - proc["Ingresos: FSS"])
    proc["Resultado: Global SPC aj. FSS"] = (proc["Resultado: Global SPC"]
                                             - proc["Ingresos: FSS"]
                                             + proc["Intereses: FSS"])
    proc["Resultado: Primario GC-BPS aj. FSS"] = (
        proc["Resultado: Primario GC-BPS"]
        - proc["Ingresos: FSS"])
    proc["Resultado: Global GC-BPS aj. FSS"] = (
        proc["Resultado: Global GC-BPS"]
        - proc["Ingresos: FSS"]
        + proc["Intereses: FSS"])

    output = proc.loc[:, fiscal_metadata[aggregation][fss]]
    metadata._set(output, area="Cuentas fiscales y deuda",
                  currency="UYU", inf_adj="No", unit="Millones",
                  seas_adj="NSA", ts_type="Flujo", cumperiods=1)

    if unit == "gdp":
        output = transform.rolling(output, periods=12, operation="sum")
        output = transform.convert_gdp(output)
    elif unit == "usd":
        output = transform.convert_usd(output)
    elif unit == "real_usd":
        output = transform.convert_real(output, start_date=start_date,
                                        end_date=end_date)
        xr = nxr.get_monthly(update_path=update_path, revise_rows=6,
                             save_path=save_path)
        output = output.divide(xr[start_date:end_date].mean()[1])
        metadata._set(output, currency="USD")
    elif unit == "real":
        output = transform.convert_real(output, start_date=start_date,
                                        end_date=end_date)
    if seas_adj in ["trend", "seas"] and unit != "gdp" and cum == 1:
        output_trend, output_seasadj = transform.decompose(output,
                                                           trading=True,
                                                           outlier=True)
        if seas_adj == "trend":
            output = output_trend
        elif seas_adj == "seas":
            output = output_seasadj

    if cum != 1:
        output = transform.rolling(output, periods=cum, operation="sum")

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        output.to_csv(full_save_path)

    return output


def labor_rate_people(seas_adj: Union[str, None] = None,
                      update_path: Union[str, PathLike, None] = None,
                      save_path: Union[str, PathLike, None] = None,
                      name: Optional[str] = None) -> pd.DataFrame:
    """
    Get labor data, both rates and persons. Allow choosing seasonal adjustment.

    Parameters
    ----------
    seas_adj : {'trend', 'seas', None}
        Whether to seasonally adjust.
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Labor market data : pd.DataFrame

    Raises
    ------
    ValueError
        If ``seas_adj`` is given an invalid keyword.

    """
    if name is None:
        name = "tfm_labor"

    if seas_adj not in ["trend", "seas", None]:
        raise ValueError("'seas_adj' can be 'trend', 'seas' or None.")

    rates = labor.get_rates(update_path=update_path, revise_rows=6,
                            save_path=save_path, force_update=False)

    if seas_adj in ["trend", "seas"]:
        trend, seasadj = transform.decompose(rates, trading=True, outlier=True)
        if seas_adj == "trend":
            rates = trend
        elif seas_adj == "seas":
            rates = seasadj

    working_age = pd.read_excel(wap_url, skiprows=7,
                                index_col=0, nrows=92).dropna(how="all")
    ages = list(range(14, 90)) + ["90 y más"]
    working_age = working_age.loc[ages].sum()
    working_age.index = pd.date_range(start="1996-06-30", end="2050-06-30",
                                      freq="A-JUN")
    monthly_working_age = working_age.resample("M").interpolate("linear")
    monthly_working_age = monthly_working_age.loc[rates.index]
    persons = rates.iloc[:, [0, 1]].div(100).mul(monthly_working_age, axis=0)
    persons["Desempleados"] = rates.iloc[:, 2].div(100).mul(persons.iloc[:, 0])
    persons.columns = ["Activos", "Empleados", "Desempleados"]
    seas_text = "NSA"
    if seas_adj == "trend":
        seas_text = "Trend"
    elif seas_adj == "seas":
        seas_text = "SA"
    metadata._set(persons, area="Mercado laboral", currency="-",
                  inf_adj="No", unit="Personas", seas_adj=seas_text,
                  ts_type="-", cumperiods=1)

    output = pd.concat([rates, persons], axis=1)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        output.to_csv(full_save_path)

    return output


def labor_real_wages(seas_adj: Union[str, None] = None,
                     update_path: Union[str, PathLike, None] = None,
                     save_path: Union[str, PathLike, None] = None,
                     name: Optional[str] = None) -> pd.DataFrame:
    """
    Get real wages. Allow choosing seasonal adjustment.

    Parameters
    ----------
    seas_adj : {'trend', 'seas', None}
        Whether to seasonally adjust.
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Real wages data : pd.DataFrame

    Raises
    ------
    ValueError
        If ``seas_adj`` is given an invalid keyword.

    """
    if name is None:
        name = "tfm_wages"

    if seas_adj not in ["trend", "seas", None]:
        raise ValueError("'seas_adj' can be 'trend', 'seas' or None.")

    wages = labor.get_wages(update_path=update_path, revise_rows=6,
                            save_path=save_path, force_update=False)
    real_wages = wages.copy()
    real_wages.columns = ["Índice medio de salarios reales",
                          "Índice medio de salarios reales privados",
                          "Índice medio de salarios reales públicos"]
    metadata._set(real_wages, area="Mercado laboral", currency="UYU",
                  inf_adj="Sí", seas_adj="NSA", ts_type="-", cumperiods=1)
    real_wages = transform.convert_real(real_wages)
    output = pd.concat([wages, real_wages], axis=1)

    if seas_adj in ["trend", "seas"]:
        trend, seasadj = transform.decompose(output,
                                             trading=True, outlier=False)
        if seas_adj == "trend":
            output = trend
        elif seas_adj == "seas":
            output = seasadj

    output = transform.base_index(output, start_date="2008-07-31")

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        output.to_csv(full_save_path)

    return output


def nat_accounts(supply: bool = True, real: bool = True, index: bool = False,
                 off_seas_adj: bool = False, usd: bool = False, cum: int = 1,
                 seas_adj: Union[str, None] = None,
                 variation: Union[str, None] = None,
                 update_path: Union[str, PathLike, None] = None,
                 save_path: Union[str, PathLike, None] = None,
                 name: Optional[str] = None) -> Optional[pd.DataFrame]:
    """Get national accounts data.

    Attempt to find one of the available data tables with the selected
    combination of parameters (``supply``, ``real``. ``index`` and
    ``seas_adj``).

    Parameters
    ----------
    supply : bool, default True
        Supply or demand side.
    real : bool, default True
        Constant or current.
    index : bool, default False
        Base 100 index or not.
    off_seas_adj : bool, default True
        Seasonally adjusted or not.
    usd : bool, default False
        If ``True``, convert to USD.
    cum : int, default 1
        How many periods to accumulate for rolling sums.
    seas_adj : {None, 'trend', 'seas'}
        Whether to seasonally adjust.
    variation : {None, 'last', 'inter', 'annual}
        Type of percentage change to calculate. Can be ``last``, ``inter`` or
        ``annual``.
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Selected national accounts : pd.DataFrame or None

    Raises
    ------
    KeyError
        If the combined parameters do not correspond to an available table.
    ValueError
        If ``seas_adj`` or ``variation`` are given invalid keywords.

    """
    if name is None:
        name = "tfm_na"

    if seas_adj not in ["trend", "seas", None]:
        raise ValueError("'seas_adj' can be 'trend', 'seas' or None.")
    if variation not in ["last", "inter", "annual", None]:
        raise ValueError("'variation' can be 'last', 'inter', 'annual', or"
                         " None.")

    data = national_accounts.get(update_path=update_path, revise_rows=4,
                                 save_path=save_path, force_update=False)

    search_terms = []
    if supply is True:
        search_terms += ["ind"]
    else:
        search_terms += ["gas"]
    if real is True:
        search_terms += ["con"]
    else:
        search_terms += ["cur"]
    if index is True:
        search_terms += ["idx"]
    if off_seas_adj is True:
        search_terms += ["sa"]
    else:
        search_terms += ["nsa"]

    table = "_".join(search_terms)

    try:
        output = data[table]
    except KeyError:
        raise KeyError("No available tables with selected parameters.")

    if usd is True:
        output = transform.convert_usd(output)

    if seas_adj in ["trend", "seas"] and off_seas_adj is False and cum == 1:
        trend, seasadj = transform.decompose(output, trading=True,
                                             outlier=True)
        if seas_adj == "trend":
            output = trend
        elif seas_adj == "seas":
            output = seasadj

    if cum != 1:
        output = transform.rolling(output, periods=cum, operation="sum")

    if variation is not None:
        output = transform.chg_diff(output, operation="chg",
                                    period_op=variation)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        output.to_csv(full_save_path)

    return output



import datetime as dt
import urllib
from os import PathLike, path, mkdir
from pathlib import Path
from typing import Union, Optional

import pandas as pd
from dateutil.relativedelta import relativedelta

from econuy.utils import metadata
from econuy.utils.lstrings import (reserves_url, reserves_cols,
                                   missing_reserves_url, ff_url)


def get(update_path: Union[str, PathLike, None] = None,
        save_path: Union[str, PathLike, None] = None,
        name: Optional[str] = None) -> pd.DataFrame:
    """Get spot, future and forwards FX operations by the Central Bank.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Daily spot, future and forwards foreign exchange operations : pd.DataFrame

    """
    if name is None:
        name = "fx_spot_ff"
    changes = _reserves_changes(update_path=update_path, save_path=save_path)
    ff = _futures_forwards(update_path=update_path, save_path=save_path)
    spot = changes.iloc[:, [0]]
    fx_ops = pd.merge(spot, ff, how="outer", left_index=True, right_index=True)
    fx_ops = fx_ops.loc[(fx_ops.index >= ff.index.min()) &
                        (fx_ops.index <= spot.index.max())]
    fx_ops = fx_ops.apply(pd.to_numeric, errors="coerce")
    fx_ops = fx_ops.fillna(0)
    fx_ops.columns = ["Spot", "Futuros", "Forwards"]

    metadata._set(fx_ops, area="Reservas internacionales",
                  currency="USD", inf_adj="No", unit="Millones",
                  seas_adj="NSA", ts_type="Flujo", cumperiods=1)
    fx_ops.columns.set_levels(["-"], level=2, inplace=True)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        fx_ops.to_csv(full_save_path)

    return fx_ops


def _reserves_changes(update_path: Union[str, PathLike, None] = None,
                      save_path: Union[str, PathLike, None] = None,
                      name: Optional[str] = None) -> pd.DataFrame:
    """Get international reserves change data.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Daily international reserves changes : pd.DataFrame

    """
    if name is None:
        name = "reserves_chg"
    months = ["ene", "feb", "mar", "abr", "may", "jun",
              "jul", "ago", "set", "oct", "nov", "dic"]
    years = list(range(2013, dt.datetime.now().year + 1))
    files = [month + str(year) for year in years for month in months]

    urls = [f"{reserves_url}{file}.xls" for file in files]
    wrong_may14 = f"{reserves_url}may2014.xls"
    fixed_may14 = f"{reserves_url}mayo2014.xls"
    urls = [fixed_may14 if x == wrong_may14 else x for x in urls]

    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        try:
            previous_data = pd.read_csv(full_update_path, index_col=0,
                                        header=list(range(9)),
                                        float_precision="high")
            metadata._set(previous_data)
            previous_data.columns.set_levels(["-"], level=2, inplace=True)
            previous_data.columns = reserves_cols[1:46]
            previous_data.index = pd.to_datetime(previous_data.index)
            urls = urls[-18:]
        except FileNotFoundError:
            previous_data = pd.DataFrame()
            pass

    reports = []
    for url in urls:

        try:
            with pd.ExcelFile(url) as xls:
                month_of_report = pd.read_excel(xls, sheet_name="INDICE")
                raw = pd.read_excel(xls, sheet_name="ACTIVOS DE RESERVA",
                                    skiprows=3)
            first_day = month_of_report.iloc[7, 4]
            last_day = (first_day
                        + relativedelta(months=1)
                        - dt.timedelta(days=1))
            proc = raw.dropna(axis=0, thresh=20).dropna(axis=1, thresh=20)
            proc = proc.transpose()
            proc.index.name = "Date"
            proc = proc.iloc[:, 1:46]
            proc.columns = reserves_cols[1:46]
            proc = proc.iloc[1:]
            proc.index = pd.to_datetime(proc.index, errors="coerce")
            proc = proc.loc[proc.index.dropna()]
            proc = proc.loc[first_day:last_day]
            reports.append(proc)

        except urllib.error.HTTPError:
            print(f"{url} could not be reached.")
            pass

    mar14 = pd.read_excel(missing_reserves_url, index_col=0)
    mar14.columns = reserves_cols[1:46]
    reserves = pd.concat(reports + [mar14], sort=False).sort_index()

    if update_path is not None:
        reserves = previous_data.append(reserves, sort=False)
        reserves = reserves.loc[~reserves.index.duplicated(keep="last")]

    reserves = reserves.apply(pd.to_numeric, errors="coerce")
    metadata._set(reserves, area="Reservas internacionales",
                  currency="USD", inf_adj="No", unit="Millones",
                  seas_adj="NSA", ts_type="Flujo", cumperiods=1)
    reserves.columns.set_levels(["-"], level=2, inplace=True)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        reserves.to_csv(full_save_path)

    return reserves


def _futures_forwards(update_path: Union[str, PathLike, None] = None,
                      save_path: Union[str, PathLike, None] = None,
                      name: Optional[str] = None) -> pd.DataFrame:
    """Get future and forwards FX operations by the Central Bank.

    Parameters
    ----------
    update_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to find a CSV
        for updating, or ``None``, don't update.
    save_path : str, os.PathLike or None, default None
        Path or path-like string pointing to a directory where to save the CSV,
        or ``None``, don't save.
    name : str, default None
        CSV filename for updating and/or saving.

    Returns
    -------
    Daily future and forwards foreign exchange operations : pd.DataFrame

    """
    if name is None:
        name = "fx_ff"
    dates = pd.bdate_range("2013-11-01",
                           dt.datetime.today()).strftime("%y%m%d").tolist()
    if update_path is not None:
        full_update_path = (Path(update_path) / name).with_suffix(".csv")
        try:
            prev_data = pd.read_csv(full_update_path, index_col=0,
                                    header=list(range(9)),
                                    float_precision="high")
            metadata._set(prev_data)
            prev_data.columns.set_levels(["-"], level=2, inplace=True)
            prev_data.index = pd.to_datetime(prev_data.index)
            last_date = prev_data.index[len(prev_data)-1]
            dates = pd.bdate_range(
                last_date + dt.timedelta(days=1), dt.datetime.today()
            ).strftime("%y%m%d").tolist()
        except FileNotFoundError:
            prev_data = pd.DataFrame()
            pass

    reports = []
    for date in dates:

        try:
            raw_report = pd.read_excel(f"{ff_url}{date}.xls")

            if (dt.datetime.strptime(date, "%y%m%d") >=
                    dt.datetime(2014, 5, 21)):
                future1 = raw_report.iloc[19, ].apply(pd.to_numeric,
                                                      errors="coerce").sum()
                future2 = raw_report.iloc[21, ].apply(pd.to_numeric,
                                                      errors="coerce").sum()
                future = future1 + future2
                forward = raw_report.iloc[24, ].apply(pd.to_numeric,
                                                      errors="coerce").sum()

            else:
                future = raw_report.iloc[19, ].apply(pd.to_numeric,
                                                     errors="coerce").sum()
                forward = raw_report.iloc[22, ].apply(pd.to_numeric,
                                                      errors="coerce").sum()

            reports.append([dt.datetime.strptime(date, "%y%m%d"),
                            future, forward])

        except urllib.error.HTTPError:
            print(f"Report for {date} could not be reached.")
            pass

    try:
        operations = pd.DataFrame(reports)
        operations.columns = ["Date", "Futuros", "Forwards"]
        operations.set_index("Date", inplace=True)
        metadata._set(
            operations, area="Reservas internacionales",  currency="USD",
            inf_adj="No", unit="Millones", seas_adj="NSA",
            ts_type="Flujo", cumperiods=1
        )
        operations.columns.set_levels(["-"], level=2, inplace=True)
        operations = operations.divide(1000)
    except ValueError:
        return prev_data

    if update_path is not None:
        operations = prev_data.append(operations, sort=False)

    if save_path is not None:
        full_save_path = (Path(save_path) / name).with_suffix(".csv")
        if not path.exists(path.dirname(full_save_path)):
            mkdir(path.dirname(full_save_path))
        operations.to_csv(full_save_path)

    return operations
